"""Charts for trend analysis."""

import typing
from collections.abc import Callable
from functools import cached_property, partialmethod

import bqplot as bq
import ipywidgets as w
import numpy as np
import pandas as pd

from ..cases import CaseSupportsChartAnaly
from ..charts import STYLE_TOOLTIP, Groups, OHLCCaseBase, tooltip_html_style
from ..config import COL_ADV, COL_DEC
from ..formatters import formatter_datetime, formatter_float, formatter_percent
from ..utils import bq_utils as ubq
from .movements import MovementBase, MovementsSupportChartAnaly, Movement, MovementAlt


class TrendsChart(OHLCCaseBase):
    """OHLC and Trend analysis for single financial instrument.

    OHLC chart with trend overlay. Trend dislayed as coloured line where
    green indicates a rising trend, red indicates a  falling trend and
    yellow indicates no trend (consolidation).

    For each movement described by `movements` a scatter mark denotes:
        movement start (triangle)
        movement end (cross)
        bar when movement start confirmed (circle)
        bar when movement end confirmed (square)

    Scatter marks for advancing movements are green, those for declining
    movements are red.

    Parameters
    ----------
    As for OHLCCaseBase, except:

    cases
        Movements representing trends over period to be charted. Must
        be passed.

    handler_click_case
        Any client-defined handler to be additionally called when user
        clicks on any scatter point representing a trend start. Should have
        signature:
            f(case: Movement, mark: bq.Scatter, event: dict)

        NB handler will be called after any handlers defined on this class.

    inc_conf_marks
        Whether to include scatter marks indicating the bar / price at
        which movements were confirmed to have started and ended.
    """

    COLOR_MAP = ["yellow", COL_ADV, COL_DEC]  # 0 consol, 1 adv, -1 dec
    MARKER_MAP: typing.ClassVar = {
        "cross": "End",
        "circle": "Start conf",
        "square": "End conf",
    }

    def __init__(
        self,
        prices: pd.DataFrame,
        title: str,
        visible_x_ticks: pd.Interval | None = None,
        max_ticks: int | None = None,
        log_scale: bool = True,
        display: bool = False,
        cases: MovementsSupportChartAnaly | None = None,
        handler_click_case: Callable | None = None,
        inc_conf_marks: bool = True,
        data_y2: list[float | int] | None = None,
    ):
        if cases is None:
            raise ValueError("'cases' is a required argument.")
        self._inc_conf_marks = inc_conf_marks

        self._SCATTER_INDEXES_ADV: list[int]  # set by _add_marks
        self._SCATTER_INDEXES_DEC: list[int]  # set by _add_marks

        super().__init__(
            prices,
            cases,
            title,
            visible_x_ticks,
            max_ticks,
            log_scale,
            display,
            handler_click_case,
            data_y2,
        )

        self.cases: MovementsSupportChartAnaly

    @cached_property
    def _y_trend(self) -> pd.Series:
        """y data for all points on trend line."""
        # set starts of advances as day low, starts of declines as day high, ends
        # of advances (that do not conincide with a subsequent trend start) as day
        # high, ends of declines (that do not conincide with a subsequent trend
        # start) as day low, otherwise as day close.
        y = self.prices.close.copy()
        subset = self.cases.starts_adv
        y[subset] = self.prices.low[subset]
        subset = self.cases.starts_dec
        y[subset] = self.prices.high[subset]
        subset = self.cases.ends_adv_solo
        y[subset] = self.prices.high[subset]
        subset = self.cases.ends_dec_solo
        y[subset] = self.prices.low[subset]
        return y

    @cached_property
    def _cols_trend(self) -> list[str]:
        """color data for all points on trend line."""
        return [self.COLOR_MAP[t] for t in self.cases.trend.values]

    def _create_scales(self) -> dict[ubq.ScaleKeys, bq.Scale]:
        scales = super()._create_scales()
        scales["width"] = bq.LinearScale()
        return scales

    create_scatter = partialmethod(OHLCCaseBase.create_scatter, opacities=[0.65])
    _add_scatter = partialmethod(OHLCCaseBase._add_scatter, opacities=[0.65])

    @property
    def current_case(self) -> MovementBase | None:
        """Currently selected case."""
        case = super().current_case
        if typing.TYPE_CHECKING:
            assert isinstance(case, MovementBase)
        return case

    def _add_marks(self, **_):
        """Add initial extra marks.

        Adds FlexLine mark to represent trend (to `Groups.PERSIST`).

        Adds Scatter marks to `Groups.CASES_SCATTERS`.
            To represent, for each trend:
                Start (triangle-up/triangle-down for adv/dec).
                End (cross) only if does not coincide with subsequent
                    trend start.

            If `inc_conf_marks` passed as True (default) to constructor
            then also adds scatter marks to represent, for each trend:
                Conf Start (circle)
                Conf End (square).
        """
        trend_line = bq.FlexLine(
            scales=self.scales,
            colors=self._cols_trend,
            x=self._x_data,
            y=self._y_trend,
            stroke_width=2,
        )
        self.add_marks([trend_line], Groups.PERSIST)

        cases = self.cases

        self._SCATTER_INDEXES_ADV = [0, 2]
        self._SCATTER_INDEXES_DEC = [1, 3]

        # add conf marks first as want them at the bottom of any pile up
        if self._inc_conf_marks:
            self._SCATTER_INDEXES_ADV += [4, 6]
            self._SCATTER_INDEXES_DEC += [5, 7]

            # scatters for confirmed starts of trends
            handler = self.handler_hover_conf_start
            self._add_scatter(
                cases.starts_conf_adv,
                cases.starts_conf_adv_px,
                COL_ADV,
                "circle",
                handler,
            )
            self._add_scatter(
                cases.starts_conf_dec,
                cases.starts_conf_dec_px,
                COL_DEC,
                "circle",
                handler,
            )
            # scatters for confirmed ends of trends
            handler = self.handler_hover_conf_end
            self._add_scatter(
                cases.ends_conf_adv,
                cases.ends_conf_adv_px,
                COL_ADV,
                "square",
                handler,
            )
            self._add_scatter(
                cases.ends_conf_dec,
                cases.ends_conf_dec_px,
                COL_DEC,
                "square",
                handler,
            )

        # scatters for ends of trends
        handler = self.handler_hover_end
        subset = cases.ends_adv
        self._add_scatter(subset, self._y_trend[subset], COL_ADV, "cross", handler)
        subset = cases.ends_dec
        self._add_scatter(subset, self._y_trend[subset], COL_DEC, "cross", handler)

        # add starts of trends last so that they appear on top of any pile
        # scatter for starts of advancing trends
        subset = cases.starts_adv
        self._add_scatter(
            subset,
            self._y_trend[subset],
            COL_ADV,
            "triangle-up",
            self.handler_hover_start,
            self._handler_click_case,
        )
        # scatter for starts of declining trends
        subset = cases.starts_dec
        self._add_scatter(
            subset,
            self._y_trend[subset],
            COL_DEC,
            "triangle-down",
            self.handler_hover_start,
            self._handler_click_case,
        )

    @property
    def mark_trend(self) -> bq.FlexLine:
        return self.added_marks[Groups.PERSIST][0]

    def update_trend_mark(self, *_):
        """Update trend mark to reflect plotted x ticks.

        See doc on base `OHLCCaseBase` class.

        Notes
        -----
        Necessary to update FlexLine to reflected plotted data to avoid the
        line doubling back over the render as it tries to plot the
        'next point after the visible range' to the start of the x-axis.

        An alternative approach, to that implemented here, to resolve the
        'wrapped around line' is to set line widths to 0 for the segments
        prior to the first and subsequent to the last plotted interval.
        This still leaves a thin trace, for which also set the colour of
        these segments to the same as the background colour. It's an
        option, although found it rendered less smoothly that the
        implemented approach (i.e. updating the marks values to reflect the
        plotted intervals).
        """
        self.mark_trend.x = self.plotted_x_ticks
        self.mark_trend.y = self._y_trend[self._domain_bv]
        self.mark_trend.colors = [
            col for col, b in zip(self._cols_trend, self._domain_bv) if b
        ]

        index = next(
            (i for i, m in enumerate(self.figure.marks) if m.name == "Flexible lines")
        )
        self.figure.marks = [m for m in self.figure.marks if m.name != "Flexible lines"]
        self.figure.marks = (
            self.figure.marks[:index] + [self.mark_trend] + self.figure.marks[index:]
        )

    def select_case(self, case: CaseSupportsChartAnaly):
        """Select a specific case."""
        case = typing.cast(MovementBase, case)
        i = self.cases.get_index_for_direction(case)
        scatter_index = 0 if case.is_adv else 1
        self._simulate_click_case(i, scatter_index)

    # HANDLERS
    def handler_hover_start(self, mark: bq.marks.Scatter, event: dict):
        """Handler for hovering on mark representing a movement start."""
        case = self.cases.event_to_case(mark, event)
        mark.tooltip.value = self.cases.get_case_html(case)

    def _handler_not_start(self, mark: bq.Scatter, event: dict):
        """Handler to hover over any scatter mark not representing a movement start.

        Displays tooltip describing what scatter represents and corresponding timestamp.
        """
        style = tooltip_html_style(color=mark.colors[0], line_height=1.3)
        date = ubq.discontinuous_date_to_timestamp(event["data"]["x"])
        name = self.MARKER_MAP[mark.marker]
        s = f"<p {style}>{name}: {formatter_datetime(date)}"
        s += f"<br>{name + ' px'}: {formatter_float(event['data']['y'])}</p"
        mark.tooltip.value = s

    def handler_hover_end(self, mark: bq.Scatter, event: dict):
        """Handler for hovering on mark representing a movement end."""
        self._handler_not_start(mark, event)

    def handler_hover_conf_start(self, mark: bq.Scatter, event: dict):
        """Handler for hovering on mark representing when movement start confirmed."""
        self._handler_not_start(mark, event)

    def handler_hover_conf_end(self, mark: bq.Scatter, event: dict):
        """Handler for hovering on mark representing when movement end confirmed."""
        self._handler_not_start(mark, event)

    def _create_line(
        self, data: pd.Series, color_: str, line_style: str, desc: str
    ) -> bq.Lines:
        """Create a `bq.Lines` mark."""
        tooltip_str = f"<p {tooltip_html_style(color=color_)}>{desc}</p>"
        line = bq.Lines(
            x=data.index,
            y=data,
            scales=self.scales,
            colors=[color_],
            line_style=line_style,
            tooltip=w.HTML(value=tooltip_str),
            tooltip_style=STYLE_TOOLTIP,
        )
        return line

    show_added_marks_all = partialmethod(OHLCCaseBase.show_added_marks_all, 0.65)
    opacate_added_marks = partialmethod(OHLCCaseBase.opacate_added_marks, opacity=0.65)

    @staticmethod
    def show_only_one_scatter_point(index: int, scatter: bq.Scatter):
        """Show only the point of a given `scatter` at a given `index`.

        Notes
        -----
        Overrides inherited method to set opacities to 0.65.
        """
        arr = np.array([0.65 if i == index else 0 for i in range(len(scatter.x))])
        scatter.opacities = arr

    def _add_case_rectangle(self, case: MovementBase, mark: bq.Scatter):
        """Add rect representing chg between conf pxs for specific case."""
        color = mark.colors[0]
        if case.closed:
            if typing.TYPE_CHECKING:
                assert case.conf_chg_pct is not None
                assert case.end_conf_px is not None
            style = tooltip_html_style(color=color)
            s = f"<p {style}>Conf chg: "
            chg_color = COL_DEC if case.conf_chg_pct < 0 else COL_ADV
            chg_style = tooltip_html_style(color=chg_color)
            s += f"<span {chg_style}>{formatter_percent(case.conf_chg_pct)}</span></p>"

            if case.is_adv:
                color_area = (
                    COL_ADV if case.start_conf_px < case.end_conf_px else COL_DEC
                )
            else:
                color_area = (
                    COL_ADV if case.start_conf_px > case.end_conf_px else COL_DEC
                )

            mark_chg = bq.Lines(
                x=[[case.start_conf, case.end_conf]] * 2,
                y=[[case.start_conf_px] * 2, [case.end_conf_px] * 2],
                scales=self.scales,
                opacities=[0],
                fill="between",
                fill_colors=[color_area],
                fill_opacities=[0.2],
                tooltip=w.HTML(value=s),
                tooltip_style=STYLE_TOOLTIP,
            )
            self.add_marks([mark_chg], Groups.CASE, under=True)

    def _add_case_marks(self, case: CaseSupportsChartAnaly, mark: bq.Scatter):
        """Add marks to represent specific case."""
        case = typing.cast(Movement, case)
        self._add_case_rectangle(case, mark)

        color_break, color_limit = "white", "slategray"
        if not case.by_break:
            color_break, color_limit = color_limit, color_break

        self.add_marks(
            [
                self._create_line(case.line_break, color_break, "dashed", "Break line"),
                self._create_line(case.line_limit, color_limit, "dashed", "Limit line"),
            ],
            Groups.CASE,
        )

    def focus_case(self, case: CaseSupportsChartAnaly, mark: bq.Scatter):
        """Focus on a given case.

        Adds scatters and lines to represent specific case.
        """
        case = typing.cast(Movement, case)
        self.remove_case_marks()
        self.reset_marks()
        idx = self.cases.get_index_for_direction(case)
        exclude_scatters = (
            self._SCATTER_INDEXES_DEC if case.is_adv else self._SCATTER_INDEXES_ADV
        )
        self.show_only_one_cases_scatter_point(idx, exclude_scatters=exclude_scatters)
        for i, mark in enumerate(self.added_marks[Groups.CASES_SCATTERS]):
            if i in exclude_scatters:
                mark.visible = False
        self._add_case_marks(case, mark)


class TrendsAltChart(TrendsChart):
    """OHLC and Trend analysis for single financial instrument.

    As TrendsChart, save for this class creating charts to visualise
    trends evaluted by the `trends.analy.TrendsAlt` class.
    """

    def _add_case_marks(self, case: CaseSupportsChartAnaly, mark: bq.Scatter):
        """Add marks to represent specific case."""
        case = typing.cast(MovementAlt, case)
        self._add_case_rectangle(case, mark)

        marks = []
        color = mark.colors[0]

        def fl(data: pd.Series, color_: str, line_style: str, desc: str) -> bq.Lines:
            line = self._create_line(data, color_, line_style, desc)
            marks.append(line)
            return line

        fl(case.sel, color, "dashed", "Start Establishment Line")
        fl(case.start_conf_line, color, "dash_dotted", "Confirmed Start Line")
        fl(
            case.end_line_consol,
            "white" if case.by_consol else "slategray",
            "dashed",
            "Conf End Line by Consolidation",
        )
        end_line_rvr_col = "white" if case.by_rvr_and_by_pct else "slategray"
        end_line_rvr = fl(
            case.end_line_rvr,
            end_line_rvr_col,
            "dash_dotted",
            "Conf End Line by Reversal",
        )

        # create labels for end_line_rvr
        rvr_arr = case.rvr_arr
        xs, ys, texts = [end_line_rvr.x[0]], [end_line_rvr.y[0]], [str(rvr_arr[0])]
        for i, (x, y, rvr_) in enumerate(
            zip(end_line_rvr.x[1:], end_line_rvr.y[1:], rvr_arr[1:])
        ):
            if rvr_ == rvr_arr[i]:
                continue
            xs.append(x)
            ys.append(y)
            texts.append(str(rvr_))
        label_mark = bq.Label(
            scales=self.scales,
            x=xs,
            y=ys,
            text=texts,
            colors=[end_line_rvr_col],
            default_size=11,
            y_offset=15 if case.is_adv else -15,
            enable_move=True,
            restrict_y=True,
            visible=False,
        )
        marks.append(label_mark)

        def end_line_rvr_handler(*_):
            label_mark.visible = not label_mark.visible

        end_line_rvr.on_element_click(end_line_rvr_handler)

        if case.end_line_rvr_opp is not None:
            fl(
                case.end_line_rvr_opp,
                "white",
                "dash_dotted",
                "Conf End Line by Opposing Movement",
            )
        if case.eel is not None:
            fl(case.eel, color, "dotted", "End Establishment Line")

        self.add_marks(marks, Groups.CASE)



"""GUIs to visualise and analyse price data.

Classes
-------
Base(metaclass=ABCMeta):
    ABC defining implementation and base functionality.

DateSliderMixin():
    Include wu.DateRangeSlider to interactively select plot dates.

BasePrice(Base):
    Base implementation for GUIs incorporating a price chart.

GuiLine(Base):
    Line price chart GUI for single financial instrument.

GuiMultLine(Base):
    Line price chart GUI for comparing multiple financial instruments.

GuiOHLC(Base):
    OHLC chart GUI for single financial instrument.

GuiPctChg(Base, DateSliderMixin):
    Bar chart GUI of precentage changes of single instrument.

GuiPctChgMult(PctChg):
    Bar chart GUI of precentage changes of multiple instruments.

Notes
-----
Module might benefit from a more compositional approach, as opposed to
hierarchical, to creating the gui interfaces. For example:

    Selector
        To select periods.

    Crosshairs
        To administer crosshairs.

    TabsControl
        UI to create crosshairs and undertake analysis.

    DataSlider
        To scross through plot dates.

Before diving in making changes, would need to have a look at the current
implementation and get clear how to best implement a compositional
approach.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Callable
from contextlib import contextmanager
from copy import deepcopy
from typing import Literal

import bqplot as bq
from bqplot.interacts import Selector
import exchange_calendars as xcals
import IPython
import ipyvuetify as v
import ipywidgets as w
import market_prices as mp
from market_prices.intervals import TDInterval, PTInterval, to_ptinterval, ONE_DAY
import pandas as pd

from market_analy import charts, gui_parts, analysis as ma_analysis
from market_analy.standalone import get_pct_off_high
from market_analy.utils.bq_utils import Crosshairs, FastIntervalSelectorDD
from market_analy.utils.dict_utils import set_kwargs_from_dflt
import market_analy.utils.ipyvuetify_utils as vu
import market_analy.utils.ipywidgets_utils as wu
import market_analy.utils.pandas_utils as upd

from .cases import CasesSupportsChartAnaly, CaseSupportsChartAnaly

# CONSTANTS
HIGHLIGHT_COLOR = "lightyellow"


class Base(metaclass=ABCMeta):
    """ABC for creating GUI.

    ABC defines:
        Implementation to create GUI
        Functionality via Mixin methods

    Notes
    -----
    Subclasses must implement the following abstract properties and methods:

    **Abstact Properties**

        ChartCls:
            charts module class, for example charts.Line.

        SelectorCls -> type[Selector] | None:
            Selector class or None if selector not required.

        _gui_box_contents -> list[w.Widget]:
            List of widgets that comprise gui.

    **Abstact Methods**

        _create_gui_parts(self):
            Method must be extended* to define all gui parts, except chart,
            with each part assigned to an instance attribute that can be
            included to `_gui_box_contents`.
            * Extension must execute method as defined on this abstract
            base class.

    Subclasses can optionally extend the following methods:

        _create_selector():
            Subclass can extend by way of passing through kwargs which will
            in turn be passed to `SelectorCls`.

        _create_gui_box():
            Subclass can extend in order to pass through kwargs which will
            in turn be passed to box `Layout` constructor.

    Subclasses should NOT override or extend the following methods:

        _create_gui()

    It is not intended that subclasses override or extend the following
        methods:

        _create_chart()

    Mixin Properties
    ----------------
    slctd_interval -> pd.Interval | None:
        Selected interval.

    Mixin Methods
    -------------
    close():
        Close all widgets that comprise gui.

    delete():
        Close and delete all widgets that comprise gui.

    display():
        Display gui.

    _show_loading_overlay():
        Show overlay indicating app is loading.

    _hide_loading_overlay():
        Hide overlay indicating app is loading.

    _popup():
        Overlay application and display popup dialog.

    _close_popup():
        Hide any popup dialog.

    _resize_chart():
        Stretch chart to fill all available horizontal space.

    _update_chart():
        Update chart with new data.

    _cycle_legend_handler():
        Handle a request to cycle chart legend to next position.
    """

    def __init__(
        self,
        data: pd.DataFrame | pd.Series,
        title: str | None = None,
        display=True,
        **kwargs,
    ):
        """Create GUI.

        Parameters
        ----------
        data
            Data to be passed to constructor of chart class returned by
            `ChartCls`.

        title
            Chart title.

        display
            True to display created GUI.

        **kwargs
            All excess kwargs will be passed to constructor of chart class
            (i.e. class as returned by `ChartCls`).
        """
        self.chart = self._create_chart(data, title=title, **kwargs)
        # Attributes set by --_create_gui--.
        # NB Furthertype[charts.Base] attributes will be set by subclass via _create_gui_parts
        self._widgets: list[w.Widget]
        self.slctr: Selector
        self._gui: w.Box

        self._create_gui()

        if display:
            self.display()

    @property
    @abstractmethod
    def ChartCls(self):
        """A class of charts module. For example, charts.Line

        Notes
        -----
        Instance of this class will instantiated by `_create_mark`.
        """

    def _create_chart(
        self, data: pd.DataFrame | pd.Series, title: str | None, **kwargs
    ) -> charts.Base:
        """Create chart.

        Notes
        -----
        Not intended that subclass extends.

        `kwargs` receive kwargs as collected by constructor.
        """
        c = self.ChartCls(data, title=title, display=False, **kwargs)
        c.figure.layout.margin = "-20px -10px -30px 0"
        return c

    @property
    @abstractmethod
    def SelectorCls(self) -> type[Selector] | None:
        """Selector class.

        Notes
        -----
        Subclass must implement. Return None if selector not required."""
        return None

    @property
    def slctd_interval(self) -> pd.Interval | None:
        """Selected interval. None if no selector or no selection."""
        if self.SelectorCls is None or self.slctr.interval is None:
            return None
        else:
            left = self.slctr.interval[0]
            right_tick = self.slctr.interval[1]
            last_interval = self.chart.date_interval_containing(right_tick)
            return pd.Interval(left, last_interval.right, closed="left")

    def _create_selector(self, **kwargs) -> type[Selector] | None:
        """Create selector.

        Notes
        -----
        Subclass can optionally extend to pass through kwargs."""
        if self.SelectorCls is None:
            return None
        kwargs.setdefault("color", "crimson")
        kwargs.setdefault("enable", False)
        kwargs["marks"] = [self.chart.mark]
        kwargs["scale"] = self.chart.scales["x"]
        kwargs["figure"] = self.chart.figure
        return self.SelectorCls(**kwargs)

    def _create_gui(self):
        """Create gui.

        Notes
        -----
        Subclass should not overwrite or extend this method."""
        self._widgets = []
        w.Widget.on_widget_constructed(lambda w: self._widgets.append(w))
        self.slctr: Selector = self._create_selector()
        self._create_gui_parts()
        self._gui: v.App = self._create_gui_box()
        w.Widget.on_widget_constructed(None)

    def _show_loading_overlay(self):
        """Show loading overlay over gui box."""
        self._loading_overlay.value = True

    def _hide_loading_overlay(self):
        """Hide loading overlay over gui box."""
        self._loading_overlay.value = False

    def _resize_chart(self):
        """Stretch chart to fill all available horizontal space."""
        old = self.chart.figure.layout.align_self
        new = "stretch" if old in [None, "flex-end"] else "flex-end"
        self.chart.figure.layout.align_self = new

    def _popup(self, title: str | None = None, text: str | None = None):
        """Overlay application and display popup dialog."""
        self._dialog.popup(title, text)

    def _close_popup(self):
        """Hide any popup dialog."""
        self._dialog.close_dialog()

    @abstractmethod
    def _create_gui_parts(self):
        """Create gui parts.

        Notes
        -----
        Subclass must extend to define all parts to contribute to gui,
        assigning each to a suitable instance attribute. Subclass must
        also execute method as defined on this abstract class.
        """
        self._loading_overlay: v.Overlay = gui_parts.loading_overlay()
        self._dialog: v.Overlay = gui_parts.Dialog()

    @property
    @abstractmethod
    def _gui_box_contents(self) -> list[w.Widget]:
        """Gui box contents

        Notes
        -----
        Subclass must implement to return list of widgets that comprise
        gui, in insertion order, ordered top to bottom. Elements should
        include `self.chart.figure`.
        """

    def _create_gui_box(self, **kwargs) -> v.App:
        """Create gui box.

        Notes
        -----
        Subclass may extend to pass through kwargs for Layout."""
        base_contents = [self._dialog, self._loading_overlay]
        contents = self._gui_box_contents + base_contents
        dflt_kwargs = {
            "d_flex": True,
            "column": True,
            "justify_center": True,
            "align_stretch": True,
            "class_": "pa-3 " + gui_parts.BG,
        }
        kwargs = set_kwargs_from_dflt(kwargs, dflt_kwargs)
        return v.App(children=contents, **kwargs)

    # MIXIN METHODS

    def close(self):
        """Closes all gui widgets."""
        self.chart.close()
        for widget in self._widgets:
            widget.close()

    def delete(self):
        """Delete all gui widgets."""
        self.close()
        self.chart.delete()
        for _ in range(len(self._widgets)):
            del self._widgets[0]

    def display(self):
        """Display gui."""
        IPython.display.display(self._gui)

    def _update_chart(self, data: pd.DataFrame | pd.Series, title: str | None = None):
        """Update chart with new data."""
        self.chart.update(data, title=title)

    def _cycle_legend_handler(
        self,
        widget: w.Widget | None = None,
        event: str | None = None,
        data: dict | None = None,
    ):
        """Handle a request to cycle chart legend to next position."""
        self.chart._cycle_legend()


class BaseVariableDates(Base):
    """Interactively select plot dates via `wu.DateRangeSlider`.

    Implementation
    --------------
    `ChartCls` must inherit from `charts.BaseSubsetDD`

    To include date slider to gui:
        include following line to `_create_gui_parts` method:
            `self.date_slider: wu.DateRangeSlider = self._create_date_slider()`
        include self.data_slider within return of `_gui_box_contents()`.

    Added properties
    ----------------
    max_ticks -> int | None
        Maximum number of x-axis ticks that will be shown by default.

    _tick_interval():
        Duration covered by each chart x-tick.

    Methods of note
    ---------------
    _set_slider_limits_to_plotted_x_ticks:
        Set date slider limits to extents of currently plotted x_ticks.

    _max_x_ticks:
        Set slider to show all plottable x_ticks.

    _create_icon_row_top:
        Create top icon row for charts with variable dates.
    """

    def __init__(
        self,
        data: pd.DataFrame | pd.Series,
        title: str | None = None,
        display=True,
        max_ticks: int | None = None,
        **kwargs,
    ):
        """Constructor.

        Parameters
        ----------
        data
            Data to be passed to constructor of chart class returned by
            `ChartCls`.

        title
            Chart title.

        max_ticks
            Maximum number of x-axis ticks that will shown by default
            (client can choose to show more via slider). None for no limit.
            (passed through to chart class).

        display
            True to display created GUI.

        **kwargs
            All excess kwargs will be passed to constructor of chart class
            (i.e. class as returned by `ChartCls`).
        """
        assert issubclass(self.ChartCls, charts.BaseSubsetDD)
        super().__init__(data, title, display, max_ticks=max_ticks, **kwargs)
        self.date_slider: wu.DateRangeSlider

    @property
    def _tick_interval(self) -> pd.Timedelta:
        """Duration covered by each x-axis timestamp."""
        return self.chart.tick_interval

    def _create_date_slider(self, **kwargs) -> wu.DateRangeSlider:
        """Create date slider.

        Notes
        -----
        Subclass can customise `wu.DateRangeSlider` by passing through
        **kwargs.
        """
        kwargs.setdefault("handle_color", HIGHLIGHT_COLOR)
        kwargs.setdefault("layout", {"margin": "0 0 15px 0"})
        ds = wu.DateRangeSlider(dates=self.chart.x_ticks, **kwargs)
        ds.interval = self.chart.plotted_interval
        ds.slider.observe(self._set_chart_x_ticks_to_slider, ["index"])
        return ds

    @property
    def _date_slider_extent(self) -> pd.Interval:
        """Interval covered by date slider.

        NB unlike `_slider.extent` includes duration of interval covered
        by last tick.
        """
        ticks_extent = self.date_slider.extent
        right = self.chart.date_interval_containing(ticks_extent.right).right
        return pd.Interval(ticks_extent.left, right, closed="left")

    @property
    def _slider(self):
        return self.date_slider.slider

    def _set_chart_x_ticks_to_slider(self, event=None):
        self.chart.plotted_x_ticks = self.date_slider.interval

    @contextmanager
    def _handler_disabled(self):
        """Undertake operation within context of disabled handler.

        Undertake an operation with context of
        `_set_chart_x_ticks_to_slider` handler being disabled.
        """
        self._slider.unobserve(self._set_chart_x_ticks_to_slider, ["index"])
        yield
        self._slider.observe(self._set_chart_x_ticks_to_slider, ["index"])

    def _set_slider_limits_to_plotted_x_ticks(self):
        """Set date slider extents to reflect currently plotted x_ticks."""
        with self._handler_disabled():
            self.date_slider.dates = self.chart.plotted_x_ticks

    def _set_slider_limits_to_all_plottable_x_ticks(self):
        """Set date slider extents to reflect all plottable x_ticks."""
        with self._handler_disabled():
            self.date_slider.dates = self.chart.x_ticks

    def _update_slider_dates(self, price_interval: pd.Timedelta):
        """Update slider dates to reflect chart x_ticks. Extents unchanged.

        price_interval
            Interval by which to extend extent beyond final timestamp, to
            cover all period represented by the final timestamp.
        """
        extent = self.date_slider.extent
        extent = pd.Interval(extent.left, extent.right + price_interval, closed="left")
        ticks = self.chart.x_ticks_subset(extent)
        with self._handler_disabled():
            self.date_slider.dates = ticks

    def _set_slider_to_plotted_x_ticks(self):
        """Set slider handles to plotted x_ticks."""
        with self._handler_disabled():
            self.date_slider.interval = self.chart.plotted_interval

    # UPDATE CHART
    def _max_x_ticks(self):
        """Show data for all plottable x_ticks."""
        self.chart.reset_x_ticks()
        self._set_slider_limits_to_all_plottable_x_ticks()

    def _update_chart(  # type: ignore[override]
        self,
        data: pd.DataFrame | pd.Series | None = None,
        data2: list[float | int | list[float | int]] | None = None,
        prices: pd.DataFrame | None = None,
        title: str | None = None,
        visible_x_ticks: pd.Interval | None = None,
        reset_slider=False,
    ):
        """Update chart with new data.

        Overrides inherited method.

        Parameters
        ----------
        data : pd.DataFrame | pd.Series | None, default: None
            data against which to update the principle mark. Pass as None
            if unchanged from current data.

        data2 : list[float | int | list[float | int]] | None, default: None
            data against which to update the secondary mark. Pass as None
            if unchanged from current data_y2.

        prices : pd.DataFrame | None
            prices on which `data` is based, or None if `data` is
            unchanged.

        reset_slider : bool, default: False
            True to set slider extents to reflect all plottable x-ticks.
            False to retain existing extents to extent possible.
        """
        prior_tick_interval = self._tick_interval
        self.chart.update(data, data2, title, visible_x_ticks)
        if reset_slider:
            self._set_slider_limits_to_all_plottable_x_ticks()
        else:
            self._update_slider_dates(prior_tick_interval)
        self._set_slider_to_plotted_x_ticks()
        if prices is not None:
            self._prices = prices

    # TOP ICON ROW
    @property
    def _icon_row_top_handlers(self) -> list[Callable]:
        return [self._max_x_ticks, self._resize_chart, self.close]

    def _icon_row_top_handler_funnel(self, widget):
        """Funnel for events triggered by clicking an icon row button."""
        handlers = self._icon_row_top_handlers
        i = self._icon_row_top.children.index(widget)
        handlers[i]()

    def _create_icon_row_top(self) -> gui_parts.IconRowTop:
        num_handlers = len(self._icon_row_top_handlers)
        funnel = [self._icon_row_top_handler_funnel] * num_handlers
        return gui_parts.IconRowTop(funnel)


class BasePrice(BaseVariableDates):
    """Base implementation for GUIs incorporating a price chart.

    GUI comprises:

    IconRow: Icons to undertake chart operations.
    Label | w.ToggleButtons: Toggle buttons for tick interval (optional)
    Label | w.ToggleButtons: Toggle buttons for drawdown period (optional)
    Figure: plot
    wu.DateRangeSlider: slider to interactively set chart dates
    HtmlOutput: Housing for html output.

    TabsControl: cursor and selector tabs.
        Cursor tab:
            + and - toggle buttons to add or delete crosshairs to the plot.
            lightbulb toggle button to make all crosshairs fully opaque or
                semi-transparent.
            trash button to delete all crosshairs.

        Selector tab. When selected, clicking chart produces a selector
        which end-user can use to define a selection of dates. Tab
        includes:
            zoom button to zoom to current selection, or currently displayed
                dates if no selection made.
            up / down arrows to evaluate maximum advance / decline over
                selected period, or period comprising all plotted dates if
                no selection made.

    Implementation
    --------------
    BaseVariableDates is substantially implemented by this base class.
    Subclasses are left with implementing the following properties:
        ChartCls -> type[charts.BasePrice]

        _chart_title -> str

    In addition, subclasses can optionally extend:

        Inherited methods as described by Base documentation.

        The following class attributes introduced by this class:
            _HAS_INTERVAL_SELECTOR : bool, default: True
                Set to False on subclass to not include the
                `w.ToggleButtons` that provides for changing tick interval.

            _HAS_DRAWDOWN : bool, default: True
                Set to False if subclass should not show drawdown data.

        The following properties introduced by this class:

            _prices_kwargs -> dict:
                Immutable chart kwargs. Subclass should extend to override
                defaults and/or add immutable kwargs specific to
                `ChartCls`. Should not include 'close_only' or
                'lose_single_symbols' keys which should instead be
                declared via the dedicated properties `_prop_close_only`
                and `_lose_single_symbol` respectively.

            _operate_kwargs -> dict:
                Kwargs to pass to pt.operate of returned price data in
                order to get data to pass to chart. Subclass can extend if
                required.

            _plot_close_only -> bool:
                Query if only close price should be passed as chart data.
                True by default. Subclass should override to return False
                if require that all prices columns (open, high, low, close,
                volume) are passed as chart data.

            _lose_single_symbol -> bool:
                Query if, before passing price data to chart, level 0
                should be stripped from columns if data is for a single
                symbol. True by default. Subclass should override to return
                False if this behaviour is not required.

    Methods
    -------
    add_crosshair():
        Add crosshair to figure.

    Properties
    ----------
    data_initial -> tuple[pd.DataFrame, list[float | list[float]] | None]:
        Copies of initial data passed to chart.

    crosshairs: list
        Crosshair objects associated with figure.
    """

    TICK_INTERVALS = [
        "3M",
        "1M",
        "5D",
        "1D",
        "4H",
        "1H",
        "30T",
        "15T",
        "5T",
        "2T",
        "1T",
    ]
    TICK_INTERVALS_PT = [to_ptinterval(ti) for ti in TICK_INTERVALS]
    _HAS_INTERVAL_SELECTOR = True
    _HAS_DRAWDOWN = True

    def __init__(
        self,
        analysis: ma_analysis.Analysis,
        interval: mp.intervals.RowInterval | None = None,
        max_ticks: int | None = None,
        log_scale: bool = True,
        display: bool = True,
        chart_kwargs: dict | None = None,
        **kwargs,
    ):
        """Create GUI.

        Parameters
        ----------
        analysis
            Analysis instance representing instruments to be plotted.

        interval
            Interval covered by an x-axis tick, as 'interval` parameter
            described by `help(analysis.prices.get)`.

        max_ticks
            Maximum number of x-axis ticks that will shown by default (client
            can choose to show more via slider). None for no limit.

        log_scale
            True to plot prices against a log scale. False to plot prices
            against a linear scale.

        display
            True to display created GUI.

        chart_kwargs
            Any kwargs to pass on to the chart class.

        **kwargs
            Period for which to plot prices. Passed as period parameters as
            described by market-prices documentation for 'PricesCls.get'
            method where 'PricesCls' is the class that was passed to
            'PricesCls' parameter of `mkt_anlaysis.Analysis` to intantiate
            `analysis`.
        """
        assert issubclass(self.ChartCls, charts.BasePrice)
        chart_kwargs = {} if chart_kwargs is None else chart_kwargs
        ptinterval = None if interval is None else to_ptinterval(interval)
        # set by _set_initial_data
        self._initial_data: tuple[pd.DataFrame, list[float | list[float]] | None]
        self._initial_prices: pd.DataFrame
        self._prices: pd.DataFrame
        self._initial_price_params_non_period_: dict
        data, data2, _ = self._set_initial_data(analysis, ptinterval, kwargs)

        self._analysis = analysis
        super().__init__(
            data=data,
            title=self._chart_title,
            max_ticks=max_ticks,
            display=display,
            log_scale=log_scale,
            data_y2=data2,
            **chart_kwargs,
        )
        self.chart.set_drawdown_presentation()

    @property
    def SelectorCls(self) -> type[FastIntervalSelectorDD]:
        return FastIntervalSelectorDD

    @property
    def _chart_title(self) -> str:
        """Chart title.

        Notes
        -----
        Implement on subclass to return chart title.
        """
        raise NotImplementedError()

    @property
    def _plot_close_only(self) -> bool:
        """Query if only close price should be passed as chart data.

        Subclass should override to return False if require all prices
        columns (open, high, low, close, volume) to be passed as chart
        data.
        """
        return True

    @property
    def _lose_single_symbol(self) -> bool:
        """Strip level 0 from columns of price data if single symbol.

        Subclass should override to return False if do not want the symbol
        level of price data to be stripped before passing to chart if data
        is for a single symbol.
        """
        return True

    @property
    def _operate_kwargs(self) -> dict[str, bool]:
        """Kwargs to pass to pt.operate of returned price data.

        Kwargs for pt.operate to get data to pass to chart.
        """
        return {
            "close_only": self._plot_close_only,
            "lose_single_symbol": self._lose_single_symbol,
        }

    @property
    def _prices_kwargs(self) -> dict:
        """Fixed kwargs to pass to `market_prices.PricesBase.get`.

        Subclass can extend to override any default values or/and extend to
        introduce further immutable kwargs specific to how `ChartCls`
        requires price data.

        Kwargs that change the structure of the returned price data should
        not be included. The following kwargs should instead be declared
        via dedicated properties:
            'close_only' - define with `_plot_close_only` property
            'lose_single_symbol' - define with `_lose_single_symbol` prop
        """
        return {}

    @property
    def prices(self) -> pd.DataFrame:
        """Copy of prices on which current chart plots based."""
        return self._prices.copy()

    @property
    def data_prices_initial(
        self,
    ) -> tuple[pd.DataFrame, list[float | list[float]] | None, pd.DataFrame]:
        """Copies of initial chart data and corresponding prices.

        Returns
        -------
        initial_data: tuple[
            pd.DataFrame,
            list[float | list[float]] | None,
            pd.DataFrame,
        ]
            [0] Data for principle mark.
            [1] Data for any mark plotted against the secondary y-axis.
            [2] Prices on which data based.
        """
        data2_ = self._initial_data[1]
        data2 = None if data2_ is None else deepcopy(data2_)
        return self._initial_data[0].copy(), data2, self._initial_prices.copy()

    def _set_initial_data(
        self,
        analysis: ma_analysis.Analysis,
        interval: mp.intervals.RowInterval | None,
        period_parameters: dict,
    ) -> tuple[pd.DataFrame, list[float | list[float]] | None, pd.DataFrame]:
        """Set initial chart data.

        Parameters
        ----------
        analysis
            'analysis' parameter as received by constructor.

        interval
            'interval' parameter as received by constructor.

        period_parameters
            **kwargs receieved by constructor.

        Notes
        -----
        Initial data is evaluated once, when this method is first called.
        Subsequent calls will return a copy of the initial data.

        The method is implemented in this way in order that prices can be
        requested by a base class earlier than they would otherwise be
        called by this `BasePrice` class. For example, price data may be
        required by a subclass to undertake analysis, such as evaluting
        trends or positions, which is then be displayed visually over a
        chart of the underlying data.
        """
        if hasattr(self, "_initial_data"):
            return self.data_prices_initial

        params = period_parameters.copy()
        params["composite"] = False
        params["interval"] = interval
        prices = analysis.prices.get(**params, **self._prices_kwargs)
        self._initial_prices = self._prices = prices
        self._initial_data = self._prices_to_chart_data(prices)
        self._set_initial_price_params_non_period(params)
        return self.data_prices_initial

    def _set_initial_price_params_non_period(self, initial_price_params: dict):
        """Set initial price params excluding those that define period and interval."""
        exclude = [
            "interval",
            "start",
            "end",
            "minutes",
            "hours",
            "days",
            "weeks",
            "months",
            "years",
        ]
        params = {k: v for k, v in initial_price_params.items() if k not in exclude}
        self._initial_price_params_non_period_ = params

    @property
    def _initial_price_params_non_period(self) -> dict:
        return self._initial_price_params_non_period_

    @staticmethod
    def _get_drawdown(
        prices: pd.DataFrame, window: int | str | None
    ) -> list[float | list[float]]:
        pct = get_pct_off_high(prices, window)
        data = upd.tolists(pct)
        if len(data) == 1 and isinstance(data[0], list):
            data = data[0]
        return data

    def _prices_to_chart_data(
        self, prices: pd.DataFrame
    ) -> tuple[pd.DataFrame, list[float | list[float]] | None]:
        """Return data to pass to chart for given price data."""
        data = prices.pt.operate(**self._operate_kwargs)
        if not self._HAS_DRAWDOWN:
            return data, None
        data2 = self._get_drawdown(prices, None)
        return data, data2

    def _get_prices(self, **kwargs):
        """Get price data."""
        return self._analysis.prices.get(**kwargs, **self._prices_kwargs)

    # TOP ICON ROW
    @property
    def _icon_row_top_handlers(self) -> list[Callable]:
        return [self._max_x_ticks, self._reset, self._resize_chart, self.close]

    # INTERVAL SELECTOR
    def _interval_selector_handler(self, change):
        self._show_loading_overlay()
        try:
            self._change_tick_interval(change["old"], change["new"])
        except ValueError as err:
            self._hide_loading_overlay()
            self._popup("Interval unavailable", err.args[0])
            self._interval_selector.set_value_unobserved(change["old"])
        else:
            self._hide_loading_overlay()

    def _create_intrvl_slctr(self) -> gui_parts.IntervalSelector:
        left = self.chart.plottable_interval.left
        right = self.chart.plottable_interval.right
        intervals = zip(self.TICK_INTERVALS, self.TICK_INTERVALS_PT)
        labels = []
        for label, pt in intervals:
            to = (
                right if isinstance(pt, TDInterval) else pt.as_offset_ms.rollback(right)
            )
            if left + pt <= to:
                labels.append(label)
        return gui_parts.IntervalSelector(
            labels, self._interval_selector_handler, self._tick_interval
        )

    # DRAWDOWN SELECTOR
    def _drawdown_selector_handler(self, change: dict):
        new, old = change["new"], change["old"]
        if not new:
            self.chart.hide_y2()
            return

        if not old:
            self.chart.show_y2()
        window = None if new == "ATH" else new
        try:
            self.set_drawdown_period(window)
        except ValueError as err:
            if window == "365D" and self._interval_selector.value != ONE_DAY:
                msg = "52wk drawdown period only available with '1D' tick interval"
            else:
                msg = err.args[0]
            self._popup("Drawdown period unavailable", msg)
            change["owner"].set_value_unobserved(old)

    def _create_drawdown_slctr(self) -> gui_parts.DrawdownSelector:
        labels = [125, 90, 60, 40, 20, 10]
        return gui_parts.DrawdownSelector(labels, self._drawdown_selector_handler)

    def _create_selector_boxes(self) -> w.HBox | None:
        if self.chart.tick_interval not in self.TICK_INTERVALS_PT:
            self._HAS_INTERVAL_SELECTOR = False

        if not self._HAS_INTERVAL_SELECTOR and not self._HAS_DRAWDOWN:
            return None

        layout = w.Layout(justify_content="flex-start", margin="15px 5px 0 0")
        style = w.widgets.widget_string.LabelStyle(font_weight="bold")
        selectors, labels = [], []
        if self._HAS_INTERVAL_SELECTOR:
            self._interval_selector = self._create_intrvl_slctr()
            selectors.append(self._interval_selector)
            labels.append(w.Label("Tick interval:", layout=layout, style=style))
        if self._HAS_DRAWDOWN:
            self._drawdown_selector = self._create_drawdown_slctr()
            selectors.append(self._drawdown_selector)
            labels.append(w.Label("Drawdown:", layout=layout, style=style))
        if len(selectors) == 1:
            return w.HBox(children=[labels[0], selectors[0]])
        return w.HBox(children=[w.VBox(children=labels), w.VBox(children=selectors)])

    # TAB CONTROL
    def _cursor_tab_reset(self):
        self.slctr.disable()

    def _selector_tab_reset(self):
        self.slctr.enable()

    def _tabs_handler(self, widget, event, data):
        if not data:
            self._cursor_tab_reset()
        elif data == 1:
            self._selector_tab_reset()

    def _trash_handler(self, widget, event, data):
        self._close_crosshairs()
        self.html_output.clear()

    def _create_tabs_control(self):
        tc = gui_parts.TabsControl()
        tc.but_lightbulb.handler_on_selecting = self.crosshairs.opaque
        tc.but_lightbulb.handler_on_deselecting = self.crosshairs.fade
        tc.but_trash.on_event("click", self._trash_handler)
        tc.but_zoom.on_event("click", self._zoom_handler)
        tc.but_arrow_up.on_event("click", self._max_adv_handler)
        tc.but_arrow_down.on_event("click", self._max_dec_handler)
        tc.on_event("change", self._tabs_handler)
        return tc

    # HTML OUTPUT
    def _create_html_output(self):
        return gui_parts.HtmlOutput()

    @property
    def html_output(self) -> gui_parts.HtmlOutput:
        return self._html_output

    # GUI BOX
    def _create_gui_parts(self):
        super()._create_gui_parts()
        self._crosshairs = Crosshairs(self.chart.figure)
        self._set_mark_handlers()
        self._icon_row_top: gui_parts.IconRowTop = self._create_icon_row_top()
        self._selector_boxes: w.HBox | None = self._create_selector_boxes()
        self.date_slider: wu.DateRangeSlider = self._create_date_slider()
        self._html_output: w.HTML = self._create_html_output()
        self.tabs_control: gui_parts.TabsControl = self._create_tabs_control()

    @property
    def _gui_box_contents(self) -> list[w.Widget]:
        contents = [self._icon_row_top]
        if self._selector_boxes is not None:
            contents += [self._selector_boxes]
        contents += [
            self.chart.figure,
            self.date_slider,
            self.tabs_control,
            self.html_output,
        ]
        return contents

    # CROSSHAIRS
    @property
    def crosshairs(self) -> Crosshairs:
        return self._crosshairs

    def _close_crosshairs(self):
        self.crosshairs.close()
        self.tabs_control.cursor_toggle.deselect()

    def _ch_delete_onclick_handler(self, crosshair):
        if self.tabs_control.cursor_toggle.selected == "minus":
            crosshair.close()
            if not self.crosshairs:
                self.tabs_control.cursor_toggle.deselect()

    def _ch_standout_onclick_handler(self, crosshair):
        if self.tabs_control.cursor_toggle.selected != "minus":
            self.crosshairs.fade()
            crosshair.opacity = 1.0

    def _ch_chgside_onclick_handler(self, crosshair):
        crosshair.change_side()

    def add_crosshair(
        self,
        draggable: bool = True,
        delible: bool = True,
        standout: bool = True,
        chgside: bool = True,
        **kwargs,
    ) -> Crosshairs.CrosshairsCh:
        """Add crosshair to plot.

        draggable
            True to allow end-sure to drag crosshairs via the cross.

        delible
            True to allow end-user to delete crosshair (in accordance with
            `_ch_delete_onclick_handler`)

        standout
            True to, on-click, make full opaque and all other crosshairs
            semi-transparent.

        chgside: bool
            True to, on-click, change side of crosshair's labels.

        **kwargs:
            passed to Crosshair.
        """
        # ensure plotted again scale associated with principle mark
        kwargs.setdefault("existing_mark", self.chart.mark)

        ch = self.crosshairs.add(draggable=draggable, **kwargs)
        reg = ch.cross.on_element_click
        if delible:
            reg(lambda _, _e: self._ch_delete_onclick_handler(ch))
        if standout:
            reg(lambda _, _e: self._ch_standout_onclick_handler(ch))
        if chgside:
            reg(lambda _, _e: self._ch_chgside_onclick_handler(ch))
        return ch

    def _add_crosshair_handler(self, mark, event):
        if self.tabs_control.cursor_toggle.selected == "plus":
            self.add_crosshair()

    def _set_mark_handlers(self):
        self.chart.mark.on_background_click(self._add_crosshair_handler)

    # ZOOM and UNZOOM
    def _zoom_to_selection(self):
        """Zoom to selection

        If no selection then zooms to x_ticks currently shown on plot.
        """
        if self.slctr.has_selection:
            self.chart.plotted_x_ticks = self.slctd_interval
        self._set_slider_limits_to_plotted_x_ticks()

    def _zoom_handler(self, widget, event, data):
        self._zoom_to_selection()
        self.slctr.reset()

    @property
    def _operation_interval(self):
        """Interval over which operation should act.

        User-selected interval or currently plotted x_ticks otherwise.
        """
        if self.slctd_interval is not None:
            return self.slctd_interval
        else:
            return self.chart.plotted_interval

    # MAX ADV/DEC
    def _get_max_chg(
        self, direction: Literal["max_adv", "max_dec"], style=False, **kwargs
    ):
        interval = self._operation_interval
        method = getattr(self._analysis, direction)
        return method(
            interval=self._tick_interval,
            start=interval.left,
            end=interval.right,
            style=style,
            **kwargs,
        )

    def _max_chg_html(self, method_name: Literal["max_adv", "max_dec"]):
        styler = self._get_max_chg(method_name, style=True)
        self.html_output.display(styler.to_html(), add_padding=4)

    def _max_chg_crosshair(
        self,
        x,
        y,
        color,
        side: Literal["lesser", "greater"],
        y_offset: int | None = None,
    ):
        label_kwargs = {"y_offset": y_offset} if y_offset is not None else {}
        self.add_crosshair(
            x=x,
            y=y,
            color=color,
            draggable=False,
            line_kwargs={"side": side},
            label_kwargs=label_kwargs,
        )

    def _max_chg_crosshairs(
        self,
        row: pd.Series,
        advance: bool,
        colors: tuple[str, str],
        y_offset: int | None = None,
    ):
        """Add crosshairs at either extreme of a defined movement.

        Adds crosshairs at the start and end of a movement represented by a
        `row` of a `pd.DataFrame` returned by `Analysis.max_adv` or
        `Analysis.max_dec`.

        Parameters
        ----------
        advance
            True if movement represents a maximum advance
            False if movements represents a maximum decline.
        """
        start = row.start if row.start.tz is None else row.start.tz_localize(None)
        y = row.low if advance else row.high
        self._max_chg_crosshair(
            x=start, y=y, color=colors[0], side="lesser", y_offset=y_offset
        )
        end = row.end if row.end.tz is None else row.end.tz_localize(None)
        y = row.high if advance else row.low
        self._max_chg_crosshair(
            x=end, y=y, color=colors[1], side="greater", y_offset=y_offset
        )
        # ugly but necessary. When the crosshairs are drawn
        # figure.interaction is set to None (don't know why). The selector is
        # thereby undrawn although retains its selection. This line ensures that
        # 1) selection reflects what's seen on screen (i.e. that there isn't a
        # selection) and 2) that a new selector remains available to the user
        # should they click on the figure.
        # Might prefer to 'fix' this by redrawing the selector after the
        # crosshairs have been added. I originally tried to decorate the
        # _max_adv and _max_dec with a context manager which stored the
        # value of .selected, forced slctr.reset() and after adding the
        # crosshairrs reassigning the stored selection to .selected. However,
        # reassigning FastIntervalSelector.selected does not trigger a redrawing
        # of the  selector rectangle to reflect the newly assigned selection.
        # with the exception of assigning None, which removes is (NB noted to
        # bqplot/issues.ipynb).
        self.slctr.reset()

    def _max_adv_crosshairs(
        self, colors=("olivedrab", "limegreen"), y_offset: int | None = None
    ):
        """Add crosshairs at either extreme of a defined advance.

        Adds crosshairs at the start and end of a movement represented by a
        `row` of a `pd.DataFrame` returned by `Analysis.max_adv`.
        """
        row = self._get_max_chg("max_adv").iloc[0]
        self._max_chg_crosshairs(
            row=row, advance=True, colors=colors, y_offset=y_offset
        )

    def _max_dec_crosshairs(
        self, colors=("orangered", "crimson"), y_offset: int | None = -25
    ):
        """Add crosshairs at either extreme of a defined decline.

        Adds crosshairs at the start and end of a movement represented by a
        `row` of a `pd.DataFrame` returned by `Analysis.max_dec`.
        """
        row = self._get_max_chg("max_dec").iloc[0]
        self._max_chg_crosshairs(
            row=row, advance=False, colors=colors, y_offset=y_offset
        )

    def _is_max_chg_available(self, method_name: Literal["max_adv", "max_dec"]) -> bool:
        if self.chart.tick_interval > ONE_DAY:
            msg = (
                f"The {method_name} function is unavailable for intervals > 1 day"
                " as the underlying data does not provide for sufficient precision."
            )
            self._popup("Function unavailable", msg)
            return False
        return True

    def _max_adv_handler(self, widget, event, data):
        if not self._is_max_chg_available("max_adv"):
            return
        self._max_chg_html("max_adv")
        self._max_adv_crosshairs()

    def _max_dec_handler(self, widget, event, data):
        if not self._is_max_chg_available("max_dec"):
            return
        self._max_chg_html("max_dec")
        self._max_dec_crosshairs()

    # UPDATE CHART
    def set_drawdown_period(self, window: int | str | None):
        """Set period over which drawdown high is evaluated."""
        dd = self._get_drawdown(self.prices, window)
        self._update_chart(data2=dd)

    def _clear_user_activity(self):
        self._close_crosshairs()
        self.html_output.clear()

    def _verify_tick_interval_valid(
        self,
        old: PTInterval,
        new: PTInterval,
        plotted_interval: pd.IntervalIndex,
        cal: xcals.ExchangeCalendar,
    ):
        """Raise ValueError if new interval greater than currently plotted range.

        Raises `ValueError` if either side of one interval at `new` would
        lie outside of `plotted_interval`.
        """
        ERROR_MSG = (
            f"Interval '{new.as_pdfreq}' unavailable: the currently plotted"
            f" range is shorther than the requested interval."
        )
        left, right = plotted_interval.left, plotted_interval.right
        if new.is_intraday:
            if old.is_intraday and (right - left < new):
                raise ValueError(ERROR_MSG)
            return  # if new intraday and old > intraday then all ok

        if old.is_intraday:
            if left in cal.opens.values:
                left = cal.minute_to_session(left)
            else:
                left = cal.minute_to_future_session(left, 1)
            right = cal.minute_to_past_session(right, 1)

        if new.is_monthly:
            right = new.as_offset_ms.rollback(right)
            if right - new < left:
                raise ValueError(ERROR_MSG)
            return

        assert new.is_daily
        if not old.is_monthly:
            if old.is_daily:
                right -= ONE_DAY  # `plotted_interval.right` is last session + one day
            if cal.session_offset(left, new.days - 1) > right:
                raise ValueError(ERROR_MSG)

    def _change_tick_interval(self, old: PTInterval, new: PTInterval):
        """Change interval of price data.

        old
            Existing price interval.

        new
            Price interval to change the chart to.
        """
        cal = self._analysis.prices.calendar_default
        self._verify_tick_interval_valid(old, new, self.chart.plotted_interval, cal)

        # evaluate new price data
        start = self.chart.plottable_interval.left
        if not new.is_intraday:
            bi = self._analysis.prices.bi_daily
        else:
            for bi in reversed(self._analysis.prices.bis_intraday):
                if new == bi or not new % bi:
                    break
        date_limit = self._analysis.prices.limits_sessions[bi][0]
        if date_limit is not None and date_limit > start:
            msg = (
                f"Prices for interval '{new.as_pdfreq}' are only available from"
                f" '{mp.helpers.fts(date_limit)}' although the earliest date that"
                f" can be plotted on the chart implies that require data"
                f" from '{mp.helpers.fts(start)}'."
            )
            raise ValueError(msg)

        end = self.chart.plottable_interval.right
        if old.is_daily:
            end -= ONE_DAY
        interval = new.as_pdfreq[:-1] if new.is_monthly else new
        try:
            prices = self._get_prices(
                interval=interval,
                start=start,
                end=end,
                **self._initial_price_params_non_period,
            )
        except Exception as err:
            raise ValueError(str(err)) from None
        data, data2 = self._prices_to_chart_data(prices)
        if data.isna().any(axis=None):
            symbols = [col for col in data.columns if data[col].isna().any()]
            msg = (
                f"Prices for interval '{new.as_pdfreq}' are not available over the"
                f" current plottable dates as no price is available over this peroid"
                f" for the following symbols: {symbols}."
            )
            raise ValueError(msg)

        # evaluate visible_x_ticks range
        left = self.chart.plotted_interval.left
        right = self.chart.plotted_interval.right
        if old.is_intraday and not new.is_intraday:
            tz = self._analysis.prices.tz_default
            if left in cal.opens.values:
                left = cal.minute_to_session(left.tz_localize(tz))
            else:
                left = cal.minute_to_future_session(left.tz_localize(tz), 1)
            right = cal.minute_to_past_session(right.tz_localize(tz), 1) + ONE_DAY
        elif not old.is_intraday and new.is_intraday:
            right -= ONE_DAY  # plotted interval right is last session + one day
            if old.is_monthly:
                left = cal.date_to_session(left, "next")
                right = cal.date_to_session(right, "previous")
            left = cal.opens[left].tz_localize(None)
            right = cal.closes[right].tz_localize(None)
        elif old.is_intraday and new.is_intraday and new > old:
            # dont' show any data to the right of the right of the last shown bar
            right -= new

        vxts = pd.Interval(left, right, closed="left")

        self._clear_user_activity()
        self._update_chart(data, data2, prices, reset_slider=True, visible_x_ticks=vxts)
        if self._HAS_DRAWDOWN:
            self._drawdown_selector.set_value_unobserved("ATH")

    def _reset_chart(self):
        """Reset initial chart."""
        data, data2, prices = self.data_prices_initial
        if isinstance(data.index, pd.IntervalIndex):
            left = data.index[0].left.tz_localize(None)
            right = data.index[-1].right.tz_localize(None)
        else:
            left, right = data.index[0], data.index[-1] + ONE_DAY
        vxts = pd.Interval(left, right, closed="left")
        self._update_chart(data, data2, prices, visible_x_ticks=vxts, reset_slider=True)
        if self._HAS_INTERVAL_SELECTOR:
            self._interval_selector.set_value_unobserved(self._tick_interval)
        if self._HAS_DRAWDOWN:
            self._drawdown_selector.set_value_unobserved("ATH")
            self.chart.show_y2()

    def _reset(self):
        self.slctr.disable()
        self._clear_user_activity()
        self.tabs_control.reset()
        self._reset_chart()


class GuiLine(BasePrice):
    """GUI to display and interact with a Line Chart."""

    @property
    def ChartCls(self) -> type[charts.Line]:
        return charts.Line

    @property
    def _chart_title(self) -> str:
        return self._analysis.symbol


class GuiMultLine(BasePrice):
    """GUI to display and interact with a Multiple Line Chart."""

    def __init__(
        self,
        analysis: ma_analysis.Compare,
        interval: mp.intervals.RowInterval | None = None,
        rebase_on_zoom: bool = True,
        max_ticks: int | None = None,
        log_scale: bool = True,
        display: bool = True,
        **kwargs,
    ):
        """Create GUI.

        Parameters
        ----------
        analysis
            Compare object representing instruments to plot.

        interval
            Interval covered by an x-axis tick, as 'interval` parameter
            described by `help(analysis.prices.get)`.

        rebase_on_zoom
            if True will rebase prices following zoom.

        max_ticks
            Maximum number of x-axis ticks that will shown by default
            (client can choose to show more via slider). None for no limit.

        log_scale
            True to plot prices against a log scale. False to plot prices
            against a linear scale.

        display
            True to display created GUI.

        **kwargs
            Period for which to plot prices. Passed as period parameters as
            described by market-prices documentation for 'PricesCls.get'
            method where 'PricesCls' is the class that was passed to
            'PricesCls' parameter of `mkt_anlaysis.Compare` to intantiate
            `analysis`.
        """
        self._rebase_on_zoom = rebase_on_zoom
        self._labels: list[str]  # set by --_set_initial_data--
        super().__init__(analysis, interval, max_ticks, log_scale, display, **kwargs)

    @property
    def ChartCls(self) -> type[charts.MultLine]:
        return charts.MultLine

    @property
    def _chart_title(self) -> str:
        return "Rebased comparison"

    @property
    def _lose_single_symbol(self) -> bool:
        return False

    @property
    def _prices_kwargs(self) -> dict:
        d = super()._prices_kwargs
        d["fill"] = "both"
        return d

    def _set_initial_data(
        self,
        analysis: ma_analysis.Analysis,
        interval: mp.intervals.RowInterval | None,
        period_parameters: dict,
    ) -> tuple[pd.DataFrame, list[float | list[float]] | None, pd.DataFrame]:
        data, data2, prices = super()._set_initial_data(
            analysis, interval, period_parameters
        )
        self._labels = list(data.columns)
        return data, data2, prices

    def _create_rebase_button(self) -> vu.IconBut:
        but = gui_parts.rebase_but(class_="ml-5 mr-5")
        but.on_event("click", self.rebase_to_first_plotted_x_tick)
        return but

    def _create_legend_button(self) -> vu.IconBut:
        but = gui_parts.legend_but(class_=" ")  # class_='mr-5'
        but.on_event("click", self._cycle_legend_handler)
        return but

    def _create_controls_box(self) -> v.Layout:
        return v.Layout(
            children=[
                self._tabs_control_container,
                self._but_rebase.tt,
                self._but_legend.tt,
            ],
            d_flex=True,
            flex_row=True,
            align_center=True,
            justify_center=True,
        )

    def _create_gui_parts(self):
        super()._create_gui_parts()
        # As tabs_control serves as a container itself, in order to avoid the
        # tabs control growing within its parent container (_controls_box)
        # necessary to set class_ to 'flex-grow-0'. However this trick doesn't
        # work by including 'flex-grow-0' directly to tabs_control, rather
        # necessary to place it within its own container.
        self._tabs_control_container = v.Layout(
            children=[self.tabs_control], class_="flex-grow-0"
        )
        self._but_rebase: vu.IconBut = self._create_rebase_button()
        self._but_legend: vu.IconBut = self._create_legend_button()
        self._controls_box: w.HBox = self._create_controls_box()

    @property
    def _gui_box_contents(self) -> list[w.Widget]:
        """Contents of gui box.

        Notes
        -----
        Subclass extends method to replace tab control with controls box
        that includes legend cycle button.
        """
        contents = super()._gui_box_contents
        index = contents.index(self.tabs_control)
        contents[index] = self._controls_box
        return contents

    def _rebase_plot_prices(self, date: pd.Timestamp):
        """Rebase plot prices such that price 100 on a given `date`.

        Closes all crosshairs which would not relate to rebased prices and,
        for same reason, any html output.
        """
        self._crosshairs.close()
        self.html_output.clear()
        self.chart.rebase(date)

    def _zoom_to_selection(self):
        """Zoom to selection

        If no selection then to x_ticks currently shown on plot.
        """
        if self.slctr.has_selection:
            self.chart.plotted_x_ticks = self.slctd_interval
        else:
            self._set_chart_x_ticks_to_slider()
        self._set_slider_limits_to_plotted_x_ticks()
        if self._rebase_on_zoom:
            self._rebase_plot_prices(self.chart.plotted_interval.left)

    def rebase_to_first_plotted_x_tick(self, widget=None, event=None, data=None):
        self._rebase_plot_prices(self.chart.plotted_interval.left)

    def _get_max_chg(
        self, direction: Literal["max_adv", "max_dec"], style=False, **kwargs
    ):
        """Get max change.

        Notes
        -----
        PITA

        Max change has to account for day high and low. Other classes
        simply call the associated `._analysis` method (max_adv or
        max_dec). This isn't an option for ChartMultLine as the plotted
        prices have been rebased and hence do not correspond with the
        non-rebased prices that `_analysis` get and uses to calculate the
        changes. Accordingly this method:
            Gets price data from analysis, to include high and low columns,
            and rebases prices so that the close columns reflect the
            plotted data and high and low columns are rebased
            proportionally (a check is whether the %age diff between the
            close and the high or low on any day of the rebased data is the
            same as the %age differences for the same day of the original
            data).

            Takes subset of these prices over the period to be operated on.

            Gets maximum change by passing this subset of the rebased data
            to `._analysis._max_chg_compare`.
        """
        interval = self._tick_interval
        prices = self._analysis.prices.get(
            interval=interval,
            include=self._labels,
            start=self.chart.x_ticks[0],
            end=self.chart.x_ticks[-1],
            fill="both",
        )
        if isinstance(prices.index, pd.IntervalIndex):
            prices.index = upd.interval_index_new_tz(prices.index, None)
        else:
            prices.index.freq = None  # `get_interval_index` will fail if CustomDay
            prices.index = mp.utils.pandas_utils.get_interval_index(prices.index, "1D")
        for symbol in self._labels:
            prices[symbol] = upd.rebase_to_cell(
                prices[symbol], row=self.chart.rebase_date, col="close"
            )
        prices = prices[prices.index.overlaps(self._operation_interval)]
        start = prices.index[0].left
        end = prices.index[-1].right
        if interval == ONE_DAY:
            prices.index = prices.index.left
            end = prices.index[-1]
        direction_ = "advance" if direction == "max_adv" else "decline"
        return self._analysis.max_chg_compare(
            direction_, prices, style, start=start, end=end
        )

    def _max_chg_crosshairs(self, advance: bool):
        """Add crosshairs at either extreme of a defined movement.

        Adds crosshairs at the start and end of a movement represented by
        the row of a `pd.DataFrame` returned by `Compare.max_chg_compare`.

        Parameters
        ----------
        advance
            True if movement represents a maximum advance
            False if it represents a maximum decline.
        """
        direction: Literal["max_adv", "max_dec"] = "max_adv" if advance else "max_dec"
        max_chgs = self._get_max_chg(direction)
        for i, label in enumerate(self.chart.mark.labels):
            offset = -10 if not i else -20
            color = self.chart.mark.colors[i]
            row = max_chgs.loc[label]
            super()._max_chg_crosshairs(
                row=row, advance=advance, colors=(color, color), y_offset=offset
            )

    def _max_adv_crosshairs(self):
        self._max_chg_crosshairs(advance=True)

    def _max_dec_crosshairs(self):
        self._max_chg_crosshairs(advance=False)

    def _legend_click_handler(self, mark, event):
        """Highlight line associated with legend label clicked.

        Also highlights any crosshairs of same color as that line - There
        is no check that crosshairs of the same color are associated in any
        way with the line.

        If line associated clicked legend label is already the only
        highlighted line then makes all lines opaque (i.e. toggle action).
        """
        index = event["data"]["index"]
        opacities = self.chart.mark.opacities
        if opacities and opacities[index] == 1 and opacities.count(1.0) == 1:
            self.chart.opaque()
        else:
            self.chart.highlight_line(index)
            self.crosshairs.highlight_color(self.chart.mark.colors[index])

    def _set_mark_handlers(self):
        super()._set_mark_handlers()
        self.chart.mark.on_legend_click(self._legend_click_handler)

    def _reset(self):
        super()._reset()
        self.chart.opaque()


class GuiOHLC(BasePrice):
    """GUI to display and interact with a OHLC Chart.

    Properties (in addition to inhertied)
    ----------
    last_selected -> pd.Series:
        Price data row corresponding with most recently clicked bar.
    """

    @property
    def ChartCls(self) -> type[charts.OHLC]:
        return charts.OHLC

    @property
    def _chart_title(self) -> str:
        return self._analysis.symbol

    @property
    def _plot_close_only(self) -> bool:
        return False

    def _create_selector(self, **kwargs) -> type[Selector] | None:
        """Create selector.

        Notes
        -----
        Overrides inherited method to avoid passing mark. Seems that
        passing mark prevents selector even being able to register
        selection against scale.

        **kwargs parameter only to maintain compatibility with base class.
        """
        kw_args = {
            "color": "white",
            "enable": False,
            "scale": self.chart.scales["x"],
            "figure": self.chart.figure,
        }
        return self.SelectorCls(**kw_args)

    def _add_crosshair_handler(self, mark: bq.OHLC, event: dict):
        if self.tabs_control.cursor_toggle.selected != "plus":
            return
        index = event["data"]["index"]
        x = self.chart.mark.x[index]

        # NOTE suspect this part will be refactored out to somewhere more general.
        higher, lower = False, False
        i = 0
        data = self.chart.mark.y
        while higher and lower or (not higher and not lower):
            i += 1
            higher = data[index][1] > data[index - i][1]
            lower = data[index][2] < data[index - i][2]
            if i == 100:
                raise StopIteration("breaking infinite loop.")
        y = data[index][1] if higher else data[index][2]

        self.add_crosshair(x=x, y=y, existing_mark=mark)

    def _display_mark_data(self, mark: bq.OHLC, event: dict):
        s = self.chart._tooltip_value(mark, event)
        self.html_output.display(s)

    @property
    def last_selected(self) -> pd.Series:
        """Price data row corresponding with most recently clicked mark."""
        return self._last_selected

    def _selected_mark_handler(self, mark: bq.OHLC, event: dict):
        self._display_mark_data(mark, event)
        i = event["data"]["index"]
        self._last_selected = self.chart.data.iloc[i]

    def _mark_handler(self, mark: bq.OHLC, event: dict):
        if self.tabs_control.cursor_toggle.selected == "plus":
            self._add_crosshair_handler(mark, event)
        else:
            self._selected_mark_handler(mark, event)

    def _set_mark_handlers(self):
        self.chart.mark.on_element_click(self._mark_handler)


class GuiPctChg(BaseVariableDates):
    """GUI for bar chart showing precentage changes of single instrument.

    GUI comprises:

        Close button: Icon to close gui.
        Figure: chart
        wu.DateRangeSlider: If there are more than 10 data points then a
            slider is included to provide for interactively selecting date
            range.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        title: str | None = None,
        display: bool = True,
        max_ticks: int | None = 10,
        direction: Literal["horizontal", "vertical"] = "vertical",
    ):
        """Create GUI.

        Parameters
        ----------
        data
            One column per symbol. Column of float values representing
            percentage change over constant frequency. Index as
            `pd.IntervalIndex` with dates representing period (of constant
            frequency) that the corresponding float values offer the
            percentage change over.

        title
            Chart title.

        display
            True to display created GUI.

        max_ticks
            Maximum number of x-axis ticks that will shown by default
            (client can choose to show more via slider). None for no limit.

        direction
            Direction of bars.
        """
        super().__init__(data, title, display, max_ticks, direction=direction)

    @property
    def ChartCls(self) -> type[charts.PctChgBar]:
        return charts.PctChgBar

    @property
    def SelectorCls(self) -> type[Selector] | None:
        return None

    def _create_date_slider(self, **_):
        return super()._create_date_slider(layout={"margin": "30px 0 15px 0"})

    def _create_gui_parts(self):
        super()._create_gui_parts()
        self._icon_row_top = self._create_icon_row_top()
        if len(self.chart.x_ticks) > self.chart.max_ticks:
            self.date_slider: wu.DateRangeSlider = self._create_date_slider()

    @property
    def _gui_box_contents(self) -> list[w.Widget]:
        contents = [self._icon_row_top, self.chart.figure]
        if hasattr(self, "date_slider"):
            contents += [self.date_slider]
        return contents


class GuiPctChgMult(GuiPctChg):
    """GUI for bar chart showing precentage changes of multiple instruments.

    GUI comprises:

        Close button: Icon to close gui.
        Figure: chart
        wu.DateRangeSlider: If there are more than 10 data points then a
            slider is offered provide for interactively selecting date
            range.
        PctChgIconRowMult: Provides for:
            Selecting chart type, either stacked or group
            Cycling legend location

    Notes
    -----
    See `charts.PctChgBarMult` documentation for known bugs concerning
    grouped bar chart.
    """

    @property
    def ChartCls(self) -> type[charts.PctChgBarMult]:
        return charts.PctChgBarMult

    def _create_icon_row(self):
        return gui_parts.PctChgIconRowMult()

    def _chart_type_handler(self, widget, event, data):
        self.chart.mark.type = self._icon_row.bar_type_tog.selected

    def _set_icon_row_handlers(self):
        handlers = [self._chart_type_handler] * 2
        self._icon_row.bar_type_tog.handlers_on_selecting = handlers
        self._icon_row.legend_cycle_but.on_event("click", self._cycle_legend_handler)

    def _create_gui_parts(self):
        super()._create_gui_parts()
        self._icon_row = self._create_icon_row()
        self._set_icon_row_handlers()

    @property
    def _gui_box_contents(self) -> list[w.Widget]:
        return super()._gui_box_contents + [self._icon_row]


class GuiOHLCCaseBase(GuiOHLC):
    """Base for analysis over OHLC for single financial instrument.

    Base class to create a gui with an OHLC chart with overlaid analysis.
    Analysis can comprise multiple 'cases', where a case could represent,
    for example, a trend or a position. gui provides user controls to
    display all classes collectively or focus on a single 'current' case
    and to navigate forwards or backwards between consecutive cases.

    Parameters
    ----------
    As base class, except:

    cases
        Cases of analysis to be displayed.

    narrow_view
        When displaying a case in 'narrow' view, the number of bars
        that should be shown before the bar representing the case's start
        and after the bar representing the case's conclusion.

    wide_view
        When displaying a case in 'wide' view, the number of bars
        that should be shown before the bar representing the case's start
        and after the bar representing the case's conclusion.

    Notes
    -----
    Subclass implementation
    -----------------------
    In addition to base clases implementation requirements, subclasses can
    optionally extend or override the following methods as required:
        _gui_handler_click_case
            Gui level handler for clicking a specific case. Alternatively
            a handler can be passed within `chart_kwargs` with the key
            'handler_click_case'.
    """

    def __init__(
        self,
        analysis: ma_analysis.Analysis,
        interval: mp.intervals.RowInterval,
        cases: CasesSupportsChartAnaly,
        max_ticks: int | None = None,
        log_scale: bool = True,
        display: bool = True,
        narrow_view: int = 10,
        wide_view: int = 10,
        chart_kwargs: dict | None = None,
        **kwargs,
    ):
        self.cases = cases
        self._narrow_view = narrow_view
        self._wide_view = wide_view
        if chart_kwargs is None:
            chart_kwargs = {}
        chart_kwargs.setdefault("cases", self.cases)
        chart_kwargs.setdefault("handler_click_case", self._gui_handler_click_case)
        super().__init__(
            analysis, interval, max_ticks, log_scale, display, chart_kwargs, **kwargs
        )

    @property
    def ChartCls(self) -> type[charts.OHLCCaseBase]:
        return charts.OHLCCaseBase

    @property
    def current_case(self) -> CaseSupportsChartAnaly | None:
        """Current selected case.

        None if no case is currently selected.
        """
        return self.chart.current_case

    def _gui_handler_click_case(
        self, case: CaseSupportsChartAnaly, mark: bq.Scatter, event: dict
    ):
        """Gui level handler for clicking a specific case.

        Lightens 'show all scatters' button to indicate option available.
        Displays tooltip to html output.

        Subclass can extend or override as required.
        """
        self.cases_controls_container.lighten_single_case()
        self.cases_controls_container.but_show_all.darken()
        html = self.cases.get_case_html(case)
        self.html_output.display(html)

    def _show_all_but_handler(self, but: vu.IconBut, event: str, data: dict):
        if but.is_light:
            self.chart.hide_cases()
            but.darken()
            return

        if self.current_case is not None:
            self.chart.deselect_current_case()
            self.cases_controls_container.darken_single_case()
        else:
            self.chart.show_cases()
        but.lighten()

    def _select_next_case_handler(self, but: vu.IconBut, event: str, data: dict):
        if but.is_dark:
            return
        self.chart.select_next_case()

    def _select_prev_case_handler(self, but: vu.IconBut, event: str, data: dict):
        if but.is_dark:
            return
        self.chart.select_previous_case()

    def _set_slider_to_current_case(self, bars: int):
        """Set slider to focus on current selected case.

        Parameters
        ----------
        bars
            Number of bars to view prior to case start and following
            case's conclusion.
        """
        if self.current_case is None:
            return
        index = self.cases.data.index
        start = max(index.get_loc(self.current_case._start) - bars, 0)
        end = self.current_case._end
        if end is None:
            stop = len(index) - 1
        else:
            stop = min(index.get_loc(end) + bars, len(index) - 1)
        self.date_slider.interval = pd.Interval(index[start], index[stop], "both")

    def _narrow_view_handler(self, but: vu.IconBut, event: str, data: dict):
        if but.is_dark:
            return
        self._set_slider_to_current_case(self._narrow_view)

    def _wide_view_handler(self, but: vu.IconBut, event: str, data: dict):
        if but.is_dark:
            return
        self._set_slider_to_current_case(self._wide_view)

    def _create_cases_controls_container(self) -> v.Layout:
        controls = gui_parts.CaseControls()
        controls.but_show_all.on_event("click", self._show_all_but_handler)
        controls.but_next.on_event("click", self._select_next_case_handler)
        controls.but_prev.on_event("click", self._select_prev_case_handler)
        controls.but_narrow.on_event("click", self._narrow_view_handler)
        controls.but_wide.on_event("click", self._wide_view_handler)
        return controls

    def _create_controls_container(self) -> v.Layout:
        self._tabs_control_container = v.Layout(
            children=[self.tabs_control], class_="d-flex justify-end mr-2"
        )
        self.cases_controls_container = self._create_cases_controls_container()
        return v.Layout(
            children=[self._tabs_control_container, self.cases_controls_container],
            class_="d-flex align-center justify-center",
        )

    def _create_gui_parts(self):
        super()._create_gui_parts()
        self._controls_container = self._create_controls_container()

    @property
    def _gui_box_contents(self) -> list[w.Widget]:
        contents = [
            self._icon_row_top,
            self.chart.figure,
            self.date_slider,
            self._controls_container,
            self.html_output,
        ]
        return contents

    def _create_date_slider(self, **kwargs):
        ds = super()._create_date_slider(**kwargs)
        ds.slider.observe(self.chart.update_trend_mark, ["index"])
        return ds

    @contextmanager
    def _handler_disabled(self):
        """Undertake an operation within context of disabled handler.

        Undertake an operation with context of slider's handlers being
        disabled.

        Notes
        -----
        Messily overrides method defined on `BaseVariableDates` subclass to
        include the additional `self.chart.update_trend_mark` handler.
        """
        self._slider.unobserve(self._set_chart_x_ticks_to_slider, ["index"])
        self._slider.unobserve(self.chart.update_trend_mark, ["index"])
        yield
        self._slider.observe(self._set_chart_x_ticks_to_slider, ["index"])
        self._slider.observe(self.chart.update_trend_mark, ["index"])

    def _max_x_ticks(self):
        """Show data for all plottable x-ticks (bars).

        Notes
        -----
        Overrides inherited method to ensure any added marks are also
        updated.
        """
        self.chart.reset_x_ticks()
        self._set_slider_limits_to_all_plottable_x_ticks()
        self.chart.update_trend_mark()



"""Utility functions and classes for lists."""

from typing import Any, Literal
from copy import copy


# DEVELOPMENT NOTE: _SelectableList
# Instead of a dedicated SelectableListMult class, Why not just aad a
# `multiple`` argument to SelectableList? Arranged as two classes to
# provide a cleaner interface that doesn't require distinguishing if a
# list being added to the selection is a single element to be added
# or list of multiple elements.
class _SelectableList(list):
    """Base class to create a list object with selectable elements.

    Subclasses should define public methods to select elements. These i
    turn should call on of the following private methods:
        _add_element_to_selection()
        _add_elements_to_selection()

    Parameters
    ----------
    elements: list
        Elements that can be selected from.

    names: list | None
        Optional names by which elements can be referenced. If passed must
        be same length as elements. If not passed elements will be assigned
        names as integers using 0-based index.

    Properties
    ----------
    has_selection:
        True if there are selected items, otherwise False.

    selected:
        Current selected items

    named_elements:
        Dictionary with keys as name and values as associated element.

    Methods
    -------
    get_element():
        Get element by name.

    is_selected_element(element):
        Check if element currently selected

    is_selected_index(index):
        Check if element referenced by index is currently selected.

    is_selected_name(name):
        Check if element referenced by name is currently selected.

    deselect_all():
        Deselect all selected elements.
    """

    def __init__(self, elements: list, names: list[str | int] | None = None):
        super().__init__(elements)
        self._names = names if names is not None else list(range(len(self)))
        assert len(elements) == len(
            self._names
        ), "'elements' and 'names' must have same length"
        self._named_elements = dict(zip(self._names, self))

        self._selected: list[Any] = []

    @property
    def named_elements(self) -> dict:
        """Dictionary with keys as name and values as associated element."""
        return self._named_elements

    def get_element(self, name: str) -> Any:
        return self.named_elements[name]

    def _subset_from_indices(self, indices: list[int]):
        return [self[i] for i in indices]

    def _subset_from_names(self, names: list) -> list[Any]:
        return [self.named_elements[name] for name in names]

    @property
    def has_selection(self) -> bool:
        return bool(self._selected)

    @property
    def selected(self) -> list[Any]:
        """Currently selected elements."""
        return self._selected

    def is_selected_element(self, element: Any) -> bool:
        """Check if element is currently selected."""
        return element in self._selected

    def is_selected_index(self, index: int) -> bool:
        """Check if element at a given index is selected."""
        return self.is_selected_element(self[index])

    def is_selected_name(self, name: str) -> bool:
        """Check if element referenced by name is selected."""
        return self.is_selected_element(self.get_element(name))

    def _add_err_msg(self, elements: Any):
        msg = (
            "'" + str(elements) + "' cannot be added to selection as it "
            "is an object or includes an object that is not an element "
            "of the selectable list. Selectable elements are " + str(self) + "."
        )
        return msg

    def _add_element_to_selection(self, element: Any):
        """Add element to selection."""
        assert element in self, self._add_err_msg(element)
        self._selected.append(element)
        return element

    def _add_elements_to_selection(
        self, elements: list, reset=False, _check_in_list=True
    ):
        """Add elements to selection.

        reset:
            True/False to replace/extend existing selection.

        _check_in_list:
            False to skip checking that passed elements are all elements of
            list.
        """
        assert isinstance(elements, list)
        if _check_in_list:
            assert all(e in self for e in elements), self._add_err_msg(elements)

        if reset:
            self._selected = elements
        else:
            # remove any element already included in selection
            elements = [e for e in elements if e not in self._selected]
            self._selected.extend(elements)

        return elements

    def _remove_elements_from_selection(self, elements: list, pass_silently=True):
        """Remove elements from selection.

        pass_silently : bool, default: True
            False: raise ValueError on trying to remove an element that was
            not in selection.

            No element is removed if error raised.
        """
        assert isinstance(elements, list)

        saved_state = copy(self._selected)
        try:
            for elem in elements:
                try:
                    self._selected.remove(elem)
                except ValueError:
                    if pass_silently:
                        pass
                    else:
                        raise
        except Exception:
            self._selected = saved_state
            raise

    def deselect_all(self):
        """Remove all elements from selection."""
        self._selected = []


class SelectableList(_SelectableList):
    """List of elements, one or zero of which can be defined as selected.

    Functionality:
        Select an element
        Get selected element
        Check if a specific element is selected
        Deselect element

    For each functionality three methods are provided that each allow for
    the element to be referenced in a particular way:
        'element' direct reference to element object,
            e.g. `selected_element`
        'name' by name associated with element
            e.g. `selected_name`
        'index' by element's index position in list
            e.g. `selected_index`

    For each functionality there is a further method, with a less verbose
    name, e.g. `selected`, which references the elemment in the default
    manner defined by passing `select_by` to the constructor as one of
    'element', 'name' or 'index'.

    Properties
    ----------
    Get selected element:
        selected
        selected_element
        selected_name
        selected_index

    Methods
    -------
    Select an element:
        select()
        select_element()
        select_name()
        select_index()

    Check if an element is selected:
        is_selected()
        is_selected_element()
        is_selected_name()
        is_selected_index()

    deselect():
        Deselect any selected element.
    """

    def __init__(
        self,
        elements: list[Any],
        names: list[str | int] | None = None,
        select_by: Literal["element", "name", "index"] = "element",
        initial_selection: Any | None = None,
    ):
        """Constructor.

        Parameters
        ----------
        elements: list
            Elements that can be selected from.

        names : list | None
            Optional names by which elements can be referenced. If passed
                must be same length as elememts. If not passed elements will
                be assigned names as integers using 0-based index.

        select_by : Literal['element', 'name', 'index'], default: 'element'
            Determines how default methods reference an element:
                'element' directly to element object
                'name' by name associated with element
                'index' by element's index position in list

        initial_selection:
            Reference to any element to be initially selected. Reference
            should be passed as either element, name or index, as
            determined by `select_by`.
        """
        super().__init__(elements, names)
        assert select_by in ["element", "name", "index"]
        self._select_by = select_by

        method_name = "select_" + select_by
        self._default_select_method = getattr(self, method_name)
        method_name = "is_selected_" + select_by
        self._default_is_selected_method = getattr(self, method_name)

        if initial_selection is not None:
            self.select(initial_selection)

    @property
    def selected_element(self) -> Any | None:
        """Selected element. None if no selection."""
        if not self.has_selection:
            return None
        return self._selected[0]

    @property
    def selected_index(self) -> int | None:
        """Index of selected element. None if no selection."""
        if not self.has_selection:
            return None
        return self.index(self.selected_element)

    @property
    def selected_name(self) -> str | int | None:
        """Name of selected element. None if no selection."""
        if not self.has_selection:
            return None
        return self._names[self.selected_index]

    @property
    def selected(self) -> Any:
        """Selected element, as default reference."""
        # NB implmentation differs from other fucntionatlity by
        # way of getting default method now rather than at time of
        # instantiation. This is because property object cannot be
        # referenced from constructor (rather returns value
        # property evaluates to)
        return getattr(self, "selected_" + self._select_by)

    def is_selected(self, reference: Any) -> bool:
        """Check if element is currently selected.

        reference:
            Element to check, referenced by default preference.
        """
        return self._default_is_selected_method(reference)

    def select_element(self, element: Any) -> Any:
        """Select element."""
        self.deselect()
        return self._add_element_to_selection(element)

    def select_index(self, index: int) -> Any:
        """Select element referenced by index."""
        return self.select_element(self[index])

    def select_name(self, name: str) -> Any:
        """Select element referenced by name."""
        return self.select_element(self.named_elements[name])

    def select(self, reference: Any) -> Any:
        """Select element referenced by default preference."""
        return self._default_select_method(reference)

    def deselect(self) -> Any:
        """Deselect any selected element."""
        super().deselect_all()


class SelectableListMult(_SelectableList):
    """List of elements, zero or any number of which can defined as selected.

    Functionality:
        Select elements
        Get selected elements
        Check if a specific element is selected
        Check if specific elements are all selected
        Deselect elements

    For most functionalities, three methods are provided that each allow
    for the element to be referenced in a particular way:
        'elements' direct reference to element object
            e.g. `selected_element`
        'names' by name associated with element
            e.g. `selected_name`
        'indices' by element's index position in list
            e.g. `selected_index`

    For each functionality there is a further method, with a less verbose
    name, e.g. `selected`, which references the elemment in the default
    manner defined by passing `select_by` to the constructor as one of
    'elements', 'names' or 'indices'.

    Properties
    ----------
    Get selected elements:
        selected
        selected_elements
        selected_names
        selected_indices

    Methods
    -------
    Select elements:
        select()
        select_elements()
        select_names()
        select_indices()

    Deselect elements:
        deselect()
        deselect_elements()
        deselect_names()
        deselect_indices()

    Check if an element is selected:
        is_selected()
        is_selected_element()
        is_selected_name()
        is_selected_index()

    Check if all of specified elements are selected:
        are_selected()
        are_selected_elements()
        are_selected_names()
        are_selected_indices()

    deselect_all():
        Deselect any selected element.
    """

    singular_map = {"elements": "element", "names": "name", "indices": "index"}

    def __init__(
        self,
        elements: list[Any],
        names: list[str | int] | None = None,
        select_by: Literal["elements", "names", "indices"] = "elements",
        initial_selection=None,
    ):
        """Constructor.

        Parameters
        ----------
        elements: list
            Elements that can be selected from.

        names : list | None
            Optional names by which elements can be referenced. If passed
                must be same length as elememts. If not passed elements will
                be assigned names as integers using 0-based index.

        select_by : Literal['elements', 'names', 'indices'], default: 'elements'
            Determines how default methods reference an element:
                'elements' directly to element objects
                'names' by names associated with elements
                'indices' by elements' index positions in list

        initial_selection:
            Reference to any elements to be initially selected. Reference
            should be passed as either elements, names or indices, as
            determined by `select_by`.
        """
        super().__init__(elements, names)
        assert select_by in ["elements", "names", "indices"]
        self._select_by = select_by

        method_name = "select_" + select_by
        self._default_select_method = getattr(self, method_name)
        method_name = "is_selected_" + self.singular_map[select_by]
        self._default_is_selected_method = getattr(self, method_name)
        method_name = "are_selected_" + select_by
        self._default_are_selected_method = getattr(self, method_name)
        method_name = "deselect_" + select_by
        self._default_deselect_method = getattr(self, method_name)

        if initial_selection is not None:
            self.select(initial_selection)

    @property
    def selected_elements(self) -> None | list[Any]:
        """Selected elements. None if no selection."""
        return None if not self.has_selection else self._selected

    @property
    def selected_indices(self) -> None | list[int]:
        """Indices of selected elements. None if no selection."""
        if not self.has_selection:
            return None
        else:
            return [self.index(e) for e in self.selected_elements]

    @property
    def selected_names(self) -> None | list[str | int]:
        """Names of selected elements. None if no selection."""
        if not self.has_selection:
            return None
        else:
            return [self._names[i] for i in self.selected_indices]

    @property
    def selected(self) -> None | list[Any]:
        """Selected elements, as default reference."""
        # NB implmentation differs from other fucntionatlity by
        # way of getting default method now rather than at time of
        # instantiation. This is because property object cannot be
        # referenced from constructor (rather returns value
        # property evaluates to)
        return getattr(self, "selected_" + self._select_by)

    def is_selected_element(self, element: Any) -> bool:
        """Check if element is currently selected."""
        return element in self._selected

    def is_selected_index(self, index: int) -> bool:
        """Check if element referenced by index is currently selected."""
        return self.is_selected_element(self[index])

    def is_selected_name(self, name: str) -> bool:
        """Check if element referenced by name is currently selected."""
        return self.is_selected_element(self.get_element(name))

    def is_selected(self, reference: Any) -> bool:
        """Check if element is currently selected.

        reference:
            Element to check, referenced by default preference.
        """
        return self._default_is_selected_method(reference)

    def are_selected_elements(self, elements: list[Any]) -> bool:
        """Check if all passed elements are selected."""
        return all(e in self.selected_elements for e in elements)

    def are_selected_indices(self, indices: list[int]) -> bool:
        """Check if all elements at given indices are selected."""
        elements = self._subset_from_indices(indices)
        return self.are_selected_elements(elements)

    def are_selected_names(self, names: list[str]) -> bool:
        """Check if all elements of given names are selected."""
        elements = self._subset_from_names(names)
        return self.are_selected_elements(elements)

    def are_selected(self, references: list[Any]) -> bool:
        """Check if all passed elements are selected.

        references
            Elements to check for selection, referenced as default
            preference.
        """
        return self._default_are_selected_method(references)

    def select_elements(
        self, elements: list, reset: bool = False, _check_in_list: bool = True
    ):
        """Select elements.

        reset
            True/False to replace/extend existing selection.

        _check_in_list : bool, default: True
            False: skip checking passed elements are all elements of list.
        """
        self._add_elements_to_selection(elements, reset, _check_in_list)

    def select_indices(self, indices: list[int], reset: bool = False):
        """Select elements referenced by indices.

        reset : bool, default: False
            True/False to replace/extend existing selection.
        """
        elements = self._subset_from_indices(indices)
        self.select_elements(elements, reset, _check_in_list=False)

    def select_names(self, names: list[str], reset: bool = False):
        """Select elements referenced by names.

        reset : bool, default: False
            True/False to replace/extend existing selection.
        """
        elements = self._subset_from_names(names)
        self.select_elements(elements, reset, _check_in_list=False)

    def select(self, references: list[Any], reset: bool = False):
        """Select elements referenced by default preference.

        reset : bool, default: False
            True/False to replace/extend existing selection.
        """
        return self._default_select_method(references, reset)

    def deselect_elements(self, elements: list[Any], pass_silently: bool = True):
        """Deselect elements.

        Parameters
        ----------
        pass_silently : bool, default: True
            False to raise ValueError on trying to remove an element that
            is not selected.
        """
        return self._remove_elements_from_selection(elements, pass_silently)

    def deselect_indices(self, indices: list[int], pass_silently: bool = True):
        """Deselect elements referenced by indices.

        Parameters
        ----------
        pass_silently : bool, default: True
            False to raise ValueError on trying to remove an element that
            is not selected.
        """
        elements = self._subset_from_indices(indices)
        self.deselect_elements(elements, pass_silently)

    def deselect_names(self, names: list[str], pass_silently: bool = True):
        """Deselect elements reference by names.

        Parameters
        ----------
        pass_silently : bool, default: True
            False to raise ValueError on trying to remove an element that
            is not selected.
        """
        elements = self._subset_from_names(names)
        self.deselect_elements(elements, pass_silently)

    def deselect(self, references: list[Any], pass_silently: bool = True):
        """Deselect elements referenced by default preference."""
        return self._default_deselect_method(references, pass_silently)



"""Market Analysis.

The market analysis package comprises the following modules:

mkt_analysis
    Price analysis of one or multiple instruments.

charts
   Classes to create bqplot figures displaying financial data.

gui_parts
    Classes to create parts of a GUI. NB Parts are independent (interaction
    between parts is the responsibility of `guis` module).

guis
    Interactive GUIs. GUIs comprise widgets and components from `gui_parts`
    module and a bqplots figure. `guis` is responsible for providing all
    interactive functionality between objects.

utils subpackage
    Various utility modules
"""

from .analysis import Analysis, Compare

__all__ = [Analysis, Compare]

__copyright__ = "Copyright (c) 2023 Marcus Read"


# Resolve version
__version__ = None

from importlib.metadata import version

try:
    # get version from installed package
    __version__ = version("market_analy")
except ImportError:
    pass

if __version__ is None:
    try:
        # if package not installed, get version as set when package built
        from ._version import version
    except Exception:
        # If package not installed and not built, leave __version__ as None
        pass
    else:
        __version__ = version

del version



"""Classes and protocols to represent trend movements."""

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from functools import cached_property

import bqplot as bq
import numpy as np
import pandas as pd

from ..cases import CaseBase, CasesBase, CasesSupportsChartAnaly, CaseSupportsChartAnaly
from ..charts import tooltip_html_style
from ..config import COL_ADV
from ..formatters import formatter_datetime, formatter_float, formatter_percent

if typing.TYPE_CHECKING:
    from .analy import TrendParams, TrendAltParams


@dataclass(frozen=True, eq=False)
class MovementBase(CaseBase, ABC, CaseSupportsChartAnaly):
    """Protocol of minimum requirements to define a trend Movement.

    Attributes
    ----------
    is_adv
        True: movement represents an advance
        False: movement represents a decline.

    start
        Bar when movement started.

    start_px
        Price when movement started.

    start_conf
        Bar when movement was confirmed as having started.

    start_px
        Price when movement was confirmed as having started.

    end
        Bar when movement ended.

    end_px
        Price when movement ended.

    end_conf
        Bar when movement was confirmed as having ended.

    end_conf_px
        Price when movement was confirmed as having ended.

    duration
        Duration of trend as number of bars
    """

    is_adv: bool
    start: pd.Timestamp
    start_px: float
    start_conf: pd.Timestamp
    start_conf_px: float
    end: pd.Timestamp | None
    end_px: float
    end_conf: pd.Timestamp | None
    end_conf_px: float | None
    duration: int

    @property
    @abstractmethod
    def open(self) -> bool:
        """Movement has not ended."""
        ...

    @property
    def closed(self) -> bool:
        """Movement has ended."""
        return not self.open

    @property
    def is_dec(self) -> bool:
        """Movement is a decline."""
        return not self.is_adv

    @property
    def trend(self) -> int:
        """Trend direction. 1 for advance, -1 for decline."""
        return 1 if self.is_adv else -1

    @property
    def _start(self) -> pd.Timestamp:
        """Bar when case considered to start."""
        return self.start

    @property
    def _end(self) -> pd.Timestamp | None:
        """Bar when case considered to have concluded."""
        return self.end_conf


@dataclass(frozen=True, eq=False)
class Movement(MovementBase, CaseSupportsChartAnaly):
    """Advance or decline identified by `Trends`.

    Attributes (in addition to those provided via `MovementBase`)
    ----------
    params
        Parameters passed to `Trends`

    line_break
        Line from movement start to first occassion that movement would end
        by way of breaking Break Line (or to end of data if movement would
        not have otherwise ended by breaking the Break Line).

        Line should 'reset' on each occasion that conditions are met for
        confirming the start of a movement.

    line_limit
        Line from movement start to first occassion that movement would end
        by way of not having extended the Limit Line (or to end of data if
        movement extended Limit Line within `prd` of end of data).

        Line should 'reset' on each occasion that the prior Limit Line was
        extended.

    by_break
        True: Movement ended by way exceeding break line.
        False: Movement ended by way of not exceeding limit line.
        None: Movemment had not ended as at the end of the data set.
    """

    params: TrendParams
    line_break: pd.Series
    line_limit: pd.Series
    by_break: bool | None

    @property
    def open(self) -> bool:
        """Movement has not ended."""
        return self.by_break is None

    @property
    def chg(self) -> float:
        """Absolute change over movement."""
        return self.end_px - self.start_px

    @property
    def chg_pct(self) -> float:
        """Percentage change over movement."""
        return self.chg / self.start_px

    @property
    def conf_chg(self) -> float | None:
        """Absolute change between start and end confirmations."""
        return (
            None if self.end_conf_px is None else self.end_conf_px - self.start_conf_px
        )

    @property
    def conf_chg_pct(self) -> float | None:
        """Percentage change between start and end confirmations."""
        return None if self.conf_chg is None else self.conf_chg / self.start_conf_px


@dataclass(frozen=True, eq=False)
class MovementAlt(MovementBase, CaseSupportsChartAnaly):
    """Advance or decline identified by `Trends`.

    Attributes (in addition to those provided via MovementProto)
    ----------
    params
        Parameters passed to `Trends`

    sel
        Start Establishment Line. Line that tracks the limit for reversals
        within the period before start_conf. Line will be comprised of
        segments, each relating to an assessed start bar. The part relating
        to the actual start bar lies to the 'right' of this bar.

    start_conf_line
        Line on which the movement was confirmed as having started. See
        `Trends` class doc.

    eel
        End Establishment Line. Line that determines if an end bar, earlier
        than that which would otherwise represent the end, is justified as
        a result of the price having failed to sufficiently extend (as
        determined by the Trends 'grad' parameter) an earlier low/high
        which would therefore be considered to better represent the
        movement's end.

        None if the movement did not end by way of consolidation or had not
        ended as at the end of the available data.

    end_line_consol
        If the price failed to cross this line within 'prd' bars then the
        movement will have been considered to have ended by way of
        consolidation.

    end_line_rvr
        If the price crosses this line then the movement will be considered
        to have ended by way of reversal as a result of having exceeded
        the reversal percentage limit.

    end_line_rvr_opp
        If the price were to cross this line then the movement would be
        considered over by way of reversal as a result of a movement in the
        opposing direction being confirmed.

        None if no opposing movement was identified before the movement
        would have otherwise ended by way of consolidation or reversal as
        a result of crossing the reversal percentage limit.

    by_consol
        True: Movement ended by way of consolidation
        False: Movement ended by way of reversal.
        None: Movemment had not ended as at the end of the available data.

    by_rvr_by_pct
        True: Movement ended by way of reversal as a result of exceeding
        the reversal percentage limit.

        False: Movement ended by other means.

    rvr_max
        Maximum reversal percentage from the bar the movement start was
        confirmed, to the bar when the movement would have been considered
        to have ended by way of consolidation (or to end of available data
        if this is earlier).

    rvr_arr
        Array of percentage reversal limits, with first item representing
        the reversal limit on the bar the movement start was confirmed.
    """

    params: TrendAltParams
    sel: pd.Series
    start_conf_line: pd.Series
    eel: pd.Series | None
    end_line_consol: pd.Series
    end_line_rvr: pd.Series
    end_line_rvr_opp: pd.Series | None
    by_consol: bool | None
    by_rvr_by_pct: bool
    rvr_max: float
    rvr_arr: np.ndarray

    @property
    def open(self) -> bool:
        """Movement has not ended."""
        return self.by_consol is None

    @property
    def by_rvr(self) -> bool | None:
        """Movement ended by way of reversal.

        None if movement open.
        """
        return None if self.open else not self.by_consol

    @property
    def by_rvr_and_by_pct(self) -> bool | None:
        """Movement ended by way of reversal exceeding reversal limit."""
        return None if self.open else self.by_rvr and self.by_rvr_by_pct

    @property
    def by_rvr_and_by_opp(self) -> bool | None:
        """Movement ended by giving way to movement in opposite direction."""
        return None if self.open else self.by_rvr and not self.by_rvr_by_pct

    @property
    def chg(self) -> float:
        """Absolute change over movement."""
        return self.end_px - self.start_px

    @property
    def chg_pct(self) -> float:
        """Percentage change over movement."""
        return self.chg / self.start_px

    @property
    def conf_chg(self) -> float | None:
        """Absolute change between start and end confirmation."""
        return None if self.open else self.end_conf_px - self.start_conf_px  # type: ignore

    @property
    def conf_chg_pct(self) -> float | None:
        """Percentage change between start and end confirmation."""
        return None if self.open else self.conf_chg / self.start_conf_px  # type: ignore

    @property
    def rvr(self) -> float | None:
        """Reversal limit percentage on bar movement confirmed as ended.

        None if movement did not end by reversal on exceeding reversal
        limit.
        """
        if not self.by_rvr_and_by_pct:
            return None
        arr = self.rvr_arr
        if len(arr) == 1:
            return arr[0]
        line = self.end_line_rvr
        if len(arr) < len(line):
            return arr[-1]
        i = line.index.get_loc(self.end_conf)
        return arr[i]


@dataclass(frozen=True)
class MovementsSupportChartAnaly(CasesSupportsChartAnaly, typing.Protocol):
    """Subprotocol for movements classes to display on a chart.

    Attributes
    ----------
    cases
        Ordered sequence of all movements identified in `data`.

    data
        OHLC data in which `cases` identified.

    interval
        Period represented by each row of `data`.
    """

    cases: Sequence[MovementBase]
    data: pd.DataFrame
    interval: pd.Timedelta | pd.DateOffset

    @property
    def advances(self) -> tuple[MovementBase, ...]:
        """Movements that represent advances."""
        ...

    @property
    def declines(self) -> tuple[MovementBase, ...]:
        """Movements that represent declines."""
        ...

    @property
    def starts_adv(self) -> pd.DatetimeIndex:
        """Start bars of advances."""
        ...

    @property
    def starts_dec(self) -> pd.DatetimeIndex:
        """Start bars of declines."""
        ...

    @property
    def ends_adv(self) -> pd.DatetimeIndex:
        """End bar of advances."""
        ...

    @property
    def ends_dec(self) -> pd.DatetimeIndex:
        """End bar of declines."""
        ...

    @property
    def ends_adv_solo(self) -> pd.DatetimeIndex:
        """End bars of advances that do not coincide with Starts.

        Bars of end of advances that do not coincide with the start bar
        of a subsequent movement (advance or decline).
        """
        ...

    @property
    def ends_dec_solo(self) -> pd.DatetimeIndex:
        """End bars of declines that do not coincide with Starts.

        Bars of end of declines that do not coincide with the start bar
        of a subsequent movement (advance or decline).
        """
        ...

    @property
    def starts_conf_adv(self) -> pd.DatetimeIndex:
        """Bars of advances when movement start confirmed."""
        ...

    @property
    def starts_conf_dec(self) -> pd.DatetimeIndex:
        """Bars of declines when movement start confirmed."""
        ...

    @property
    def starts_conf_adv_px(self) -> list[float]:
        """Prices when advance movements confirmed."""
        ...

    @property
    def starts_conf_dec_px(self) -> list[float]:
        """Prices when decline movements confirmed."""
        ...

    @property
    def ends_conf_adv(self) -> pd.DatetimeIndex:
        """Bars of advances when movement end confirmed."""
        ...

    @property
    def ends_conf_dec(self) -> pd.DatetimeIndex:
        """Bars of declines when movement end confirmed."""
        ...

    @property
    def ends_conf_adv_px(self) -> list[float]:
        """Prices when end of advance movements confirmed."""
        ...

    @property
    def ends_conf_dec_px(self) -> list[float]:
        """Prices when end of decline movements confirmed."""
        ...

    @property
    def trend(self) -> pd.Series:
        """Trend.

        `pd.Series` with index as bars (`pd.DatetimeIndex`) and values as
        corresponding trend value:
            1 advance
            0 consolidation
            -1 decline
        """
        ...

    def get_index_for_direction(self, case: MovementBase) -> int:
        """Get index position of a case in `advances` or `declines`"""
        ...


@dataclass(frozen=True)
class MovementsBase(CasesBase):
    """Movements identified over an analysis period.

    Fulfills `MovementsSupportChartAnaly` with exception of handlers which
    should be added by the subclass if it is intended that analysis will
    be charted.

    Attributes
    ----------
    cases
        Ordered sequence of all movements identified in `data`.

    data
        OHLC data in which `cases` identified.

    interval
        Period represented by each row of `data`.
    """

    cases: Sequence[MovementBase]
    data: pd.DataFrame
    interval: pd.Timedelta | pd.DateOffset

    def moves(self) -> Sequence[MovementBase]:
        """Alias for `cases` to maintain backwards compatibility."""
        return self.cases

    @cached_property
    def advances(self) -> tuple[MovementBase, ...]:
        """Movements that represent advances."""
        lst = [move for move in self.cases if move.is_adv]
        return tuple(lst)

    @cached_property
    def declines(self) -> tuple[MovementBase, ...]:
        """Movements that represent declines."""
        lst = [move for move in self.cases if move.is_dec]
        return tuple(lst)

    @property
    def starts_adv(self) -> pd.DatetimeIndex:
        """Start bars of advances."""
        return pd.DatetimeIndex([move.start for move in self.advances])

    @property
    def starts_dec(self) -> pd.DatetimeIndex:
        """Start bars of declines."""
        return pd.DatetimeIndex([move.start for move in self.declines])

    @property
    def starts_conf_adv(self) -> pd.DatetimeIndex:
        """Bar of advances when movement start confirmed."""
        return pd.DatetimeIndex([move.start_conf for move in self.advances])

    @property
    def starts_conf_dec(self) -> pd.DatetimeIndex:
        """Bar of declines when movement start confirmed."""
        return pd.DatetimeIndex([move.start_conf for move in self.declines])

    @property
    def starts_conf_adv_px(self) -> list[float]:
        """Prices when advance movements confirmed."""
        return [move.start_conf_px for move in self.advances]

    @property
    def starts_conf_dec_px(self) -> list[float]:
        """Prices when decline movements confirmed."""
        return [move.start_conf_px for move in self.declines]

    @property
    def ends_adv(self) -> pd.DatetimeIndex:
        """End bar of advances."""
        return pd.DatetimeIndex([move.end for move in self.advances if move.closed])

    @property
    def ends_dec(self) -> pd.DatetimeIndex:
        """End bar of declines."""
        return pd.DatetimeIndex([move.end for move in self.declines if move.closed])

    @property
    def ends_conf_adv(self) -> pd.DatetimeIndex:
        """Bar of advances when movement end confirmed."""
        return pd.DatetimeIndex(
            [move.end_conf for move in self.advances if move.closed]
        )

    @property
    def ends_conf_dec(self) -> pd.DatetimeIndex:
        """Bar of declines when movement end confirmed."""
        return pd.DatetimeIndex(
            [move.end_conf for move in self.declines if move.closed]
        )

    @property
    def ends_conf_adv_px(self) -> list[float]:
        """Prices when end of advance movements confirmed."""
        return [
            move.end_conf_px for move in self.advances if move.end_conf_px is not None
        ]

    @property
    def ends_conf_dec_px(self) -> list[float]:
        """Prices when end of decline movements confirmed."""
        return [
            move.end_conf_px for move in self.declines if move.end_conf_px is not None
        ]

    @property
    def starts(self) -> pd.DatetimeIndex:
        """Start bar of all movements."""
        return pd.DatetimeIndex([move.start for move in self.cases])

    @property
    def ends(self) -> pd.DatetimeIndex:
        """End bar of all movements."""
        return pd.DatetimeIndex([move.end for move in self.cases if move.closed])

    @property
    def ends_adv_solo(self) -> pd.DatetimeIndex:
        """End bar of advances that do not coincide with `starts`.

        Bar of end of advances that do not coincide with the start bar of a
        subsequent movement (advance or decline).
        """
        return self.ends_adv.difference(self.starts)

    @property
    def ends_dec_solo(self) -> pd.DatetimeIndex:
        """End bar of declines that do not coincide with `starts`.

        Bar of end of declines that do not coincide with the start bar of a
        subsequent movement (advance or decline).
        """
        return self.ends_dec.difference(self.starts)

    @property
    def trend(self) -> pd.Series:
        """Trend

        `pd.Series` with index as bars (`pd.DatetimeIndex`) and values as
        corresponding trend value:
            1 advance
            0 consolidation
            -1 decline
        """
        srs = pd.Series(0, index=self.data.index)
        for move in self.cases:
            stop = None if move.end is None else move.end - self.interval
            srs[move.start : stop] += move.trend
        return srs

    def get_index_for_direction(self, case: MovementBase) -> int:
        """Get index position of a case in `advances` or `declines`"""
        seq = self.advances if case.is_adv else self.declines
        return seq.index(case)

    def event_to_case(self, mark: bq.Scatter, event: dict) -> MovementBase:
        """Get case corresonding to an event for mark representing a case.

        Parameters as those passed to an event handler.
        """
        i = event["data"]["index"]
        case = self.advances[i] if mark.colors[0] == COL_ADV else self.declines[i]
        return case


@dataclass(frozen=True)
class Movements(MovementsBase, MovementsSupportChartAnaly):
    """Movements identified over an analysis period by `Trends`."""

    cases: Sequence[Movement]

    def event_to_case(self, mark: bq.Scatter, event: dict) -> Movement:
        # Only for type checker to know move is a Movement as defined on this module.
        case = super().event_to_case(mark, event)
        case = typing.cast(Movement, case)
        return case

    @staticmethod
    def get_case_html(case: CaseSupportsChartAnaly) -> str:
        """Return html to describe a movement."""
        case = typing.cast(Movement, case)
        color = "crimson" if case.is_dec else "limegreen"
        style = tooltip_html_style(color=color, line_height=1.3)
        s = f"<p {style}>Start: " + formatter_datetime(case.start)
        s += f"<br>Start px: {formatter_float(case.start_px)}"
        s += f"<br>End: {'None' if case.end is None else formatter_datetime(case.end)}"
        s += f"<br>Chg: {formatter_percent(case.chg_pct)}"
        by = "None" if case.open else ("break" if case.by_break else "limit")
        s += f"<br>By: {by}"
        s += f"<br>Duration: {case.duration}"
        s += f"<br>Start conf: {formatter_datetime(case.start_conf)}"
        end_conf = (
            "None" if case.end_conf is None else formatter_datetime(case.end_conf)
        )
        s += f"<br>End conf: {end_conf}"

        s += "<br>Conf chg: "
        if case.open:
            s += "None"
        else:
            if typing.TYPE_CHECKING:
                assert case.conf_chg_pct is not None
            chg_color = "crimson" if case.conf_chg_pct < 0 else "limegreen"
            chg_style = tooltip_html_style(color=chg_color, line_height=1.3)
            s += f"<span {chg_style}>{formatter_percent(case.conf_chg_pct)}</span>"

        s += "</p"
        return s


@dataclass(frozen=True)
class MovementsAlt(MovementsBase, MovementsSupportChartAnaly):
    """Movements identified over an analysis period by `Trends`."""

    cases: Sequence[MovementAlt]

    def event_to_case(self, mark: bq.Scatter, event: dict) -> MovementAlt:
        # Only for type checker to know move is a Movement as defined on this module.
        case = super().event_to_case(mark, event)
        if typing.TYPE_CHECKING:
            assert isinstance(case, MovementAlt)
        return case

    @staticmethod
    def get_case_html(case: CaseSupportsChartAnaly) -> str:
        """Return html to describe a movement."""
        case = typing.cast(MovementAlt, case)
        color = "crimson" if case.is_dec else "limegreen"
        style = tooltip_html_style(color=color, line_height=1.3)
        s = f"<p {style}>Start: " + formatter_datetime(case.start)
        s += f"<br>Start px: {formatter_float(case.start_px)}"
        s += f"<br>End: {'None' if case.end is None else formatter_datetime(case.end)}"
        s += f"<br>Chg: {formatter_percent(case.chg_pct)}"
        by = "None" if case.open else ("consol" if case.by_consol else "reversal")
        s += f"<br>By: {by}"
        s += f"<br>Duration: {case.duration}"
        if case.rvr is not None:
            s += f"<br>Rrv: {formatter_percent(case.rvr)}"
        s += f"<br>Rrv Max: {formatter_percent(case.rvr_max)}"
        s += f"<br>Start conf: {formatter_datetime(case.start_conf)}"
        end_conf = (
            "None" if case.end_conf is None else formatter_datetime(case.end_conf)
        )
        s += f"<br>End conf: {end_conf}"

        s += "<br>Conf chg: "
        if case.conf_chg_pct is None:
            s += "None"
        else:
            chg_color = "crimson" if case.conf_chg_pct < 0 else "limegreen"  # type: ignore
            chg_style = tooltip_html_style(color=chg_color, line_height=1.3)
            s += f"<span {chg_style}>{formatter_percent(case.conf_chg_pct)}</span>"

        s += "</p"
        return s
